'''
Experimental task base classes, contains mostly code to run the generic
finite state machine representing different phases of the task
'''

import traceback
import collections
import re
import os
import tables
import traceback
import numpy as np
from collections import OrderedDict

from config.rig_defaults import rig_settings
from . import traits
from .. import fsm
from ..fsm import FSMTable, StateTransitions, ThreadedFSM

try:
    os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
    import pygame
    pygame_avail = True
except ImportError:
    import warnings
    warnings.warn("experiment.py: Cannot import 'pygame'")
    pygame_avail = False


min_per_hour = 60.
sec_per_min = 60.

def time_to_string(sec):
    '''
    Convert a time in seconds to a string of format hh:mm:ss.
    '''
    nhours = int(sec/(min_per_hour * sec_per_min))
    nmins = int((sec-nhours*min_per_hour*sec_per_min)/sec_per_min)
    nsecs = int(sec - nhours*min_per_hour*sec_per_min - nmins*sec_per_min)
    return str(nhours).zfill(2) + ':' + str(nmins).zfill(2) + ':' + str(nsecs).zfill(2)

def _get_trait_default(trait):
    '''
    Function which tries to determine the default value for a trait in the class declaration
    '''
    _, default = trait.default_value()
    if isinstance(default, tuple) and len(default) > 0:
        try:
            func, args, _ = default
            default = func(*args)
        except:
            pass
    return default

class ExperimentMeta(type(traits.HasTraits)):
    '''
    Metaclass that merges traits and controls across class which
    inheritance from multiple parents
    '''
    def __new__(meta, name, bases, dct):
        '''
        Merge the parent trait lists into the class trait lists
        '''
        exclude_parent_traits = set()
        ordered_traits = set()
        hidden_traits = set()

        all_dct = [cls.__dict__ for cls in bases]
        all_dct.append(dct)
        for parent in all_dct:
            for key, value in parent.items():
                if key == 'hidden_traits':
                    hidden_traits.update(value)
                elif key == 'ordered_traits':
                    ordered_traits.update(value)
                elif key == 'exclude_parent_traits':
                    exclude_parent_traits.update(value)
        dct['exclude_parent_traits'] = exclude_parent_traits
        dct['ordered_traits'] = ordered_traits
        dct['hidden_traits'] = hidden_traits
        return super().__new__(meta, name, bases, dct)

    @property
    def controls(cls):
        '''
        Lookup the methods which are tagged with bmi3d_control.
        '''
        controls = set()
        for parent in cls.__mro__:
            for key, value in parent.__dict__.items():
                if hasattr(value, 'bmi3d_control') and value.bmi3d_control:
                    controls.add(value)
        return controls

def control_decorator(fn):
    fn.bmi3d_control = True
    return fn

class Experiment(ThreadedFSM, traits.HasTraits, metaclass=ExperimentMeta):
    '''
    Common ancestor of all task/experiment classes
    '''
    status = dict(
        wait = dict(start_trial="trial", premature="penalty", stop=None),
        trial = dict(correct="reward", incorrect="penalty", timeout="penalty"),
        reward = dict(post_reward="wait", end_state=True),
        penalty = dict(post_penalty="wait", end_state=True),
    )

    # Set the initial state to 'wait'. The 'wait' state has special behavior for the Sequence class (see below)
    state = "wait"

    # Flag to indicate that the task object has not been constructed or initialized
    _task_init_complete = False

    # Flag to set in order to stop the FSM gracefully
    stop = False

    # Rate at which FSM is called. Set to 60 Hz by default to match the typical monitor update rate
    fps = traits.Float(60, desc="Rate at which the FSM is called") # Hz

    cycle_count = 0

    # set this flag to true if certain things should only happen in debugging mode
    debug = False
    terminated_in_error = False

    ## GUI/database-related attributes
    # Flag to specify if you want to be able to create a BMI Decoder object from the web interface
    is_bmi_seed = False

    # Trait GUI manipulation
    exclude_parent_traits = [] # List of possible parent traits that you don't want to be set from the web interface
    ordered_traits = [] # Traits in this list appear in order at the top of the web interface parameters
    hidden_traits = []  # These traits are hidden on the web interface, and can be displayed by clicking the 'Show' radiobutton on the web interface

    # Runtime settable traits
    session_length = traits.Float(0, desc="Time until task automatically stops. Length of 0 means no auto stop.")
    num_rewarded_trials = traits.Float(0, desc="Number of succesful trials. Length of 0 means no auto stop for succesful trials.")


    # Initialization functions -----------------------------------------------
    @classmethod
    def pre_init(cls, **kwargs):
        '''
        Jobs to do before creating the task object go here (or this method should be overridden in child classes).
        Examples might include sending a trigger to start a recording device (e.g., neural system), since you might want
        recording to be guaranteed to start before any task event loop activity occurs.
        '''
        print('running experiment.Experiment.pre_init')
        pass

    def __init__(self, verbose=True, **kwargs):
        '''
        Constructor for Experiment. This is the standard python object constructor

        Parameters
        ----------
        kwargs: optional keyword-arguments
            Any user-specified parameters for experiment traits, to be passed to the traits.HasTraits parent.

        Returns
        -------
        Experiment instance
        '''
        traits.HasTraits.__init__(self, **kwargs)
        ThreadedFSM.__init__(self)
        self.verbose = verbose
        self.task_start_time = self.get_time()
        self.reward_count = 0
        self.saveid = kwargs['saveid'] if 'saveid' in kwargs else None
        self.reportstats = collections.OrderedDict()
        self.reportstats['State'] = None # State stat is automatically updated for all experiment classes
        self.reportstats['Runtime'] = '' # Runtime stat is automatically updated for all experiment classes
        self.reportstats['Trial #'] = 0 # Trial # stat must be updated by individual experiment classes
        # If the FSM is set up in the old style (explicit dictionaries instead of wrapper data types), convert to the newer FSMTable
        if isinstance(self.status, dict):
            self.status = FSMTable.construct_from_dict(self.status)

        # Attribute for task entry dtype, used to create a numpy record array which is updated every iteration of the FSM
        # See http://docs.scipy.org/doc/numpy/user/basics.rec.html for details on how to create a record array dtype
        self.dtype = []

        self.cycle_count = 0
        if pygame_avail:
            self.clock = pygame.time.Clock()
        else:
            self.clock = fsm.Clock()

        self.pause = False


        ## Figure out which traits to not save to the HDF file
        ## Large/complex python objects cannot be saved as HDF file attributes
        ctraits = self.class_traits()
        self.object_trait_names = [ctr for ctr in list(ctraits.keys()) if ctraits[ctr].trait_type.__class__.__name__ in ['Instance', 'InstanceFromDB', 'DataFile']]

        if self.verbose: print("finished executing Experiment.__init__")

    def init(self):
        '''
        Initialization method to run *after* object construction (see self.start).
        This may be necessary in some cases where features are used with multiple inheritance to extend tasks
        (this is the standard way of creating custom base experiment + features classes through the browser interface).
        With multiple inheritance, it's difficult/annoying to make guarantees about the order of operations for
        each of the individual __init__ functions from each of the parents. Instead, this function runs after all the
        __init__ functions have finished running if any subsequent initialization is necessary before the main event loop
        can execute properly. Examples include initialization of the decoder state/parameters.
        '''
        # Timestamp for rough loop timing
        self.last_time = self.get_time()
        self.cycle_count = 0

        # Create task_data record array
        # NOTE: all data variables MUST be declared prior to this point. So child classes overriding the 'init' method must
        # declare their variables using the 'add_dtype' function BEFORE calling the 'super' method.
        try:
            self.dtype = np.dtype(self.dtype)
            self.task_data = np.zeros((1,), dtype=self.dtype)
        except:
            print("Error creating the task_data record array")
            traceback.print_exc()
            print(self.dtype)
            self.task_data = None

        # Register the "task" source with the sinks
        if not hasattr(self, 'sinks'): # this attribute might be set in one of the other 'init' functions from other inherited classes
            from riglib import sink
            self.sinks = sink.SinkManager.get_instance()

        try:
            self.sinks.register("task", self.dtype, include_msgs=True)
        except:
            traceback.print_exc()
            raise Exception("Error registering task source")

        self._task_init_complete = True

    def add_dtype(self, name, dtype, shape):
        '''
        Add to the dtype of the task. The task's dtype attribute is used to determine
        which attributes to save to file.
        '''
        new_field = (name, dtype, shape)
        existing_field_names = [x[0] for x in self.dtype]
        if name in existing_field_names:
            raise Exception("Duplicate add_dtype functionc call for task data field: %s" % name)
        else:
            self.dtype.append(new_field)

    def screen_init(self):
        '''
        This method is implemented by the riglib.stereo_opengl.Window class, which is not used by all tasks. However,
        since Experiment is the ancestor of all tasks, a stub function is here so that any children
        using the window can safely use 'super'.
        '''
        pass

    def sync_event(self, event_name, event_data=None, immediate=False):
        '''
        Stub function for sending sync signals to various devices. Could be digital triggers to a recording system
        or markers on a screen measured by photodiode, for example. Implemented in features.sync_features
        '''
        pass 

        # TODO warning that you're not sending sync signals!

    # Trait functions --------------------------------------------------------
    @classmethod
    def class_editable_traits(cls):
        '''
        Class method to retrieve the list of editable traits for the given experiment.
        The default behavior for an experiment class is to make all traits editable except for those
        listed in the attribute 'exclude_parent_traits'.

        Parameters
        ----------
        None

        Returns
        -------
        editable_traits: list of strings
            Names of traits which are designated to be runtime-editable
        '''
        # traits = super(Experiment, cls).class_editable_traits()
        from traits.trait_base import not_event, not_false
        traits = cls.class_trait_names(type=not_event, editable=not_false)
        editable_traits = [x for x in traits if x not in cls.exclude_parent_traits]
        return editable_traits

    @classmethod
    def get_trait_info(cls, trait_name, ctraits=None):
        """Get dictionary of information on a given trait"""
        if ctraits is None:
            ctraits = cls.class_traits()

        trait_params = dict()
        trait_params['type'] = ctraits[trait_name].trait_type.__class__.__name__
        trait_params['default'] = _get_trait_default(ctraits[trait_name])
        trait_params['desc'] = ctraits[trait_name].desc
        trait_params['hidden'] = 'hidden' if cls.is_hidden(trait_name) else 'visible'
        if hasattr(ctraits[trait_name], 'label'):
            trait_params['label'] = ctraits[trait_name].label
        else:
            trait_params['label'] = trait_name

        if trait_params['type'] == "InstanceFromDB":
            # a database instance. pass back the model and the query parameters and let the db
            # handle the rest
            trait_params['options'] = (ctraits[trait_name].bmi3d_db_model, ctraits[trait_name].bmi3d_query_kwargs)

        elif trait_params['type'] == 'Instance':
            raise ValueError("You should use the 'InstanceFromDB' trait instead of the 'Instance' trait!")

        elif trait_params['type'] == "Enum":
            raise ValueError("You should use the 'OptionsList' trait instead of the 'Enum' trait!")

        elif trait_params['type'] == "OptionsList":
            trait_params['options'] = ctraits[trait_name].bmi3d_input_options

        elif trait_params['type'] == "DataFile":
            trait_params['options'] = ("DataFile", ctraits[trait_name].bmi3d_query_kwargs)

        return trait_params

    @classmethod
    def get_params(cls):
        # Use an ordered dict so that params actually stay in the order they're added, instead of random (hash) order
        params = OrderedDict()

        ctraits = cls.class_traits()

        # add all the traits that are explicitly instructed to appear at the top of the menu
        ordered_traits = cls.ordered_traits
        for trait in ordered_traits:
            if trait in cls.class_editable_traits():
                params[trait] = cls.get_trait_info(trait, ctraits=ctraits)

        # add all the remaining non-hidden traits
        for trait in cls.class_editable_traits():
            if trait not in params and not cls.is_hidden(trait):
                params[trait] = cls.get_trait_info(trait, ctraits=ctraits)

        # add any hidden traits
        for trait in cls.class_editable_traits():
            if trait not in params:
                params[trait] = cls.get_trait_info(trait, ctraits=ctraits)
        return params

    def get_trait_values(self):
        '''
        Retrieve all the values of the 'trait' objects
        '''
        trait_values = dict()
        for trait in self.class_editable_traits():
            trait_values[trait] = getattr(self, trait)
        return trait_values

    @classmethod
    def is_hidden(cls, trait):
        '''
        Return true if the given trait is not meant to be shown on the GUI by default, i.e. hidden

        Parameters
        ----------
        trait: string
            Name of trait to check

        Returns
        -------
        bool
        '''
        return trait in cls.hidden_traits

    @classmethod
    def get_desc(cls, params, report):
        return "An experiment!"

    # FSM functions ----------------------------------------------------------
    def run(self):
        '''
        Method to run the finite state machine of the task. Code that needs to execute
        imediately before the task starts running in child classes should be of the form:

        def run(self):
            do stuff
            try:
                super(class_name, self).run()
            finally:
                clean up stuff

        The try block may or may not be necessary. For example, if you're opening a UDP port, you may want to always
        close the socket whether or not the main loop executes properly so that you don't loose the
        reference to the socket.
        '''

        ## Initialize the FSM before the loop
        self.screen_init()
        self.reportstats['State'] = self.state
        super(Experiment, self).run()

    def _cycle(self):
        '''
        Code that needs to run every task loop iteration goes here
        '''
        super(Experiment, self)._cycle()

        # Send task data to any registered sinks
        if hasattr(self, 'task_data') and self.task_data is not None:
            self.sinks.send("task", self.task_data)

        # Update report stats periodically
        if self.cycle_count % self.fps == 0: 
            self.update_report_stats()

    def set_state(self, condition, *args, **kwargs):
        self.reportstats['State'] = condition or 'stopped'
        super().set_state(condition, *args, **kwargs)

    def _test_stop(self, ts):
        '''
        FSM 'test' function. Returns the 'stop' attribute of the task. Will only be
        called if the current state is 'stoppable', i.e. it has a 'stop' entry in its 
        state transition table.
        '''
        if self.session_length > 0 and (self.get_time() - self.task_start_time) > self.session_length:
            self.end_task()
        if self.num_rewarded_trials > 0 and self.reward_count >= self.num_rewarded_trials:
            self.end_task()
        return self.stop

    def _test_time_expired(self, ts):
        '''
        Generic function to test if time has expired. For a state 'wait', the function looks up the
        variable 'wait_time' and uses that as a time.
        '''
        state_time_var_name = self.state + '_time'
        try:
            state_time = getattr(self, state_time_var_name)
        except AttributeError:
            raise AttributeError("Cannot find attribute %s; may not be able to use generic time_expired event for state %s"  % (state_time_var_name, self.state))

        assert isinstance(state_time, (float, int))
        return ts > state_time

    # UI interaction functions -----------------------------------------------
    @classmethod
    def _time_to_string(cls, sec):
        '''
        Convert a time in seconds to a string of format hh:mm:ss.
        '''
        nhours = int(sec/(min_per_hour * sec_per_min))
        nmins = int((sec-nhours*min_per_hour*sec_per_min)/sec_per_min)
        nsecs = int(sec - nhours*min_per_hour*sec_per_min - nmins*sec_per_min)
        return str(nhours).zfill(2) + ':' + str(nmins).zfill(2) + ':' + str(nsecs).zfill(2)

    def update_report_stats(self):
        '''
        Function to update any relevant report stats for the task. Values are saved in self.reportstats,
        an ordered dictionary. Keys are strings that will be displayed as the label for the stat in the web interface,
        values can be numbers or strings. Called on every state change.
        '''
        self.reportstats['Runtime'] = self._time_to_string(self.get_time() - self.task_start_time)

    def online_report(self):
        return self.reportstats

    def get_state(self):
        return self.state

    @classmethod
    def offline_report(self, event_log):
        '''Returns an ordered dict with report stats to be displayed when past session of this task is selected
        in the web interface. Not called while task is running, only offline, so stats must come from information
        available in a sessions event log. Inputs are task object and event_log.'''
        offline_report = collections.OrderedDict()
        if len(event_log) == 0:
            explength = 0
        else:
            explength = event_log[-1][-1] - event_log[0][-1]
        offline_report['Runtime'] = self._time_to_string(explength)
        n_trials = 0
        n_success_trials = 0
        n_error_trials = 0
        for k, (state, event, t) in enumerate(event_log):
            if state == "reward":
                n_trials += 1
                n_success_trials += 1
            elif re.match('.*?_penalty', state):
                n_trials += 1
                n_error_trials += 1
        offline_report['Total trials'] = n_trials
        offline_report['Total rewards'] = n_success_trials
        try:
            offline_report['Rewards/min'] = np.round((n_success_trials/explength) * 60, decimals=2)
        except:
            offline_report['Rewards/min'] = 0
        if n_trials == 0:
            offline_report['Success rate'] = None
        else:
            offline_report['Success rate'] = str(np.round(float(n_success_trials)/n_trials*100,decimals=2)) + '%'
        return offline_report

    @staticmethod
    def log_summary(event_log):
        '''Return summary of trials in this experimental block
        exp_log: sequence of state transitions recorded by LogExperiment class
        '''
        report = dict()

        if len(event_log) == 0:
            report['runtime'] = 0
        else:
            report['runtime'] = event_log[-1][-1] - event_log[0][-1]

        n_trials = 0
        n_success_trials = 0
        n_error_trials = 0
        for k, (state, event, t) in enumerate(event_log):
            if state == "reward":
                n_trials += 1
                n_success_trials += 1
            elif re.match('.*?_penalty', state):
                n_trials += 1
                n_error_trials += 1

        report['n_trials'] = n_trials
        report['n_success_trials'] = n_success_trials
        return report

    @staticmethod
    def format_log_summary(report):
        '''Pretty-print the output of `log_summary`'''
        offline_report = collections.OrderedDict()
        offline_report['Runtime'] = time_to_string(report['runtime'])
        offline_report['Total trials'] = report['n_trials']
        offline_report['Total rewards'] = report['n_success_trials']

        # derived metrics
        n_success_trials = float(report['n_success_trials'])
        n_trials = float(report['n_trials'])
        if report['runtime'] > 0:
            rewards_per_min = np.round(n_success_trials/report['runtime'] * 60, decimals=2)
        else:
            rewards_per_min = 0

        if n_trials > 0:
            success_rate = np.round(n_success_trials/n_trials*100, decimals=2)
        else:
            success_rate = 0

        offline_report['Rewards/min'] = rewards_per_min
        offline_report['Success rate'] = '%g %%' % success_rate
        return offline_report

    # UI cleanup functions ---------------------------------------------------
    def cleanup(self, database, saveid, **kwargs):
        '''
        Commands to execute at the end of a task if it is to be saved to the database.

        Parameters
        ----------
        database : object
            Needs to have the methods save_bmi, save_data, etc. For instance, the db.tracker.dbq module or an RPC representation of the database
        saveid : int
            TaskEntry database record id to link files/data to
        kwargs : optional dict arguments
            Optional arguments to dbq methods. NOTE: kwargs cannot be used when 'database' is an RPC object.

        Returns
        -------
        success : bool
            False will trigger an error in the web GUI
        '''
        if self.verbose: print("experimient.Experiment.cleanup executing")
        return True

    def cleanup_hdf(self):
        '''
        Method for adding data to hdf file after hdf sink is closed by
        system at end of task. The HDF file is re-opened and any extra task
        data kept in RAM is written
        '''
        if hasattr(self, "h5file"):
            traits = self.class_editable_traits()
            h5file = tables.open_file(self.h5file.name, mode='a')
            
            # Create an empty task database if there isn't one already (may be empty if there was no task data)
            if not hasattr(h5file.root, "task"):
                h5file.create_table("/", "task", [('time', 'u8')])

            # Append to task data metadata
            for trait in traits:
                if (trait not in self.object_trait_names): # don't save traits which are complicated python objects to the HDF file
                    h5file.root.task.attrs[trait] = getattr(self, trait) 
            h5file.close()

    def terminate(self):
        '''
        Cleanup commands for all tasks regardless of whether they are saved or not
        '''
        pass

    # Web-facing controls --------------------------------------
    @control_decorator
    def play_pause(self):
        self.pause = not self.pause

        if 'pause' not in self.status.states:
            self.sync_event("PAUSE", immediate=True)

        if self.pause:
            print("Paused!")
        else:
            print("...resuming!")

        return self.pause


class LogExperiment(Experiment):
    '''
    Extension of the experiment class which logs state transitions
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reportstats['Success rate'] = "0 %"
        self.reportstats['Success rate (10 trials)'] = "0 %"

    def update_report_stats(self):
        super().update_report_stats()
        n_rewards = self.calc_state_occurrences('reward')
        n_trials = max(1, self.calc_trial_num())
        self.reward_count = n_rewards
        self.reportstats['Trial #'] = n_trials - 1
        self.reportstats['Success rate'] = "{} %".format(int(100*n_rewards/n_trials))
        self.reportstats['Success rate (10 trials)'] = "{} %".format(int(100*self.calc_events_per_trial("reward", 10)))

    def cleanup(self, database, saveid, **kwargs):
        '''
        Commands to execute at the end of a task.
        Save the task event log to the database

        see riglib.Experiment.cleanup for argument descriptions
        '''
        if self.verbose: print("experiment.LogExperiment.cleanup")
        super(LogExperiment, self).cleanup(database, saveid, **kwargs)
        dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'

        report_stats = self.log_summary(self.event_log)
        print("save_log", report_stats)

        if dbname == 'default':
            database.save_log(saveid, report_stats)
        else:
            database.save_log(saveid, report_stats, dbname=dbname)
        return True
    ##########################################################
    ##### Functions to calculate statistics from the log #####
    ##########################################################
    def calc_state_occurrences(self, state_name):
        '''
        Calculate the number of times the task enters a particular state

        Parameters
        ----------
        state_name: string
            Name of state to track

        Returns
        -------
        Counts of state occurrences
        '''
        times = np.array([state[1] for state in self.state_log if state[0] == state_name])
        return len(times)

    def calc_trial_num(self):
        '''
        Counts the number of trials which have finished.
        '''
        trialtimes = [state[1] for state in self.state_log if state[0] in self.status.trial_end_states]
        return len(trialtimes)

    def calc_events_per_min(self, event_name, window):
        '''
        Calculates the rate of event_name, per minute

        Parameters
        ----------
        event_name: string
            Name of state representing "event"
        window: float
            Number of seconds into the past to look to calculate the current event rate estimate.

        Returns
        -------
        rate : float
            Rate of specified event, per minute
        '''
        times = np.array([state[1] for state in self.state_log if state[0]==event_name])
        if (self.get_time() - self.task_start_time) < window:
            divideby = (self.get_time() - self.task_start_time)/sec_per_min
        else:
            divideby = window/sec_per_min
        return np.sum(times >= (self.get_time() - window))/divideby

    def calc_time_since_last_event(self, event_name):
        '''
        Calculates the time elapsed since the previous instance of event_name
        '''
        start_time = self.state_log[0][1]
        times = np.array([state[1] for state in self.state_log if state[0]==event_name])
        if len(times):
            return times[-1] - start_time
        else:
            return np.float64("0.0")

    def calc_events_per_trial(self, event_name, window):
        '''
        Calculates the rate of event_name, per trial

        Parameters
        ----------
        event_name: string
            Name of the state to calculate
        window: int
            Number of trials into the past to calculate the rate estimate
        
        Returns
        -------
        rate : float
            Rate of specified event, per trial
        '''
        trialtimes = [state[1] for state in self.state_log if state[0] in self.status.trial_end_states]
        if len(trialtimes) == 0:
            return 0
        elif len(trialtimes) < window:
            times = np.array([state[1] for state in self.state_log if state[0]==event_name and state[1] > trialtimes[0]])
            return len(times) / max(1, len(trialtimes) - 1)
        else:
            times = np.array([state[1] for state in self.state_log if state[0]==event_name and state[1] > trialtimes[-window]])
            return  len(times) / (window - 1)

class Sequence(LogExperiment):
    '''
    Task where the targets or other information relevant to the start of each trial
    are presented by a Python generator
    '''

    # List of staticmethods of the class which can be used to create a sequence of targets for each trial
    sequence_generators = []

    @classmethod
    def get_default_seq_generator(cls):
        '''
        Define a default sequence generator as the first one listed in the 'sequence_generators' attribute
        '''
        return getattr(cls, cls.sequence_generators[0])

    def __init__(self, gen=None, *args, **kwargs):
        '''
        Constructor for Sequence

        Parameters
        ----------
        gen : Python generator
            Object with a 'next' attribute used in the special "wait" state to get the target sequence for the next trial.
        kwargs: optonal keyword-arguments
            Passed to the super constructor

        Returns
        -------
        Sequence instance
        '''
        if gen is None:
            raise ValueError("Experiment classes which inherit from Sequence must specify a target generator!")

        if hasattr(gen, '__next__'): # is iterable already
            self.gen = gen
        elif np.iterable(gen):
            from .generate import runseq
            self.gen = runseq(self, seq=gen)
        else:
            raise ValueError("Input argument to Sequence 'gen' must be of 'generator' type!")

        self.trial_dtype = np.dtype([('time', 'u8'), ('trial', 'u4')]) # to be overridden in init()

        super(Sequence, self).__init__(*args, **kwargs)

    def init(self, *args, **kwargs):

        # Create a record array for trials
        if not hasattr(self, 'sinks'): # this attribute might be set in one of the other 'init' functions from other inherited classes
            from riglib import sink
            self.sinks = sink.SinkManager.get_instance()
        dtype = self.trial_dtype # if you want to change this, do it in init() before calling super().init()
        self.trial_record = np.zeros((1,), dtype=dtype)
        self.sinks.register("trials", dtype)
        super().init(*args, **kwargs)

    def _start_wait(self):
        '''
        At the start of the wait state, the generator (self.gen) is querried for
        new information needed to start the trial. If the generator runs out, the task stops.
        '''
        if self.debug:
            print("_start_wait")

        try:
            self.next_trial = next(self.gen)
            # self._parse_next_trial() # KP changed 10/26/22
        except StopIteration:
            self.end_task()

        self._parse_next_trial()

    def _parse_next_trial(self):
        '''
        Interpret the data coming from the generator. If the generator yields a dictionary,
        then the keys of the dictionary automatically get set as attributes.

        Over-ride or add additional code in child classes if different behavior is desired.
        '''
        if isinstance(self.next_trial, dict):
            for key in self.next_trial:
                setattr(self, '_gen_%s' % key, self.next_trial[key])

        self.trial_record['time'] = self.cycle_count
        self.trial_record['trial'] = self.calc_trial_num()
        self.sinks.send("trials", self.trial_record)


class TrialTypes(Sequence):
    '''
    This module is deprecated, used by some older tasks (dots, rds)
    '''
    trial_types = []

    status = dict(
        wait = dict(start_trial="picktrial", premature="penalty", stop=None),
        reward = dict(post_reward="wait"),
        penalty = dict(post_penalty="wait"),
    )

    def __init__(self, gen, **kwargs):
        super(TrialTypes, self).__init__(gen, **kwargs)
        assert len(self.trial_types) > 0

        for ttype in self.trial_types:
            self.status[ttype] = {
                "%s_correct"%ttype :"reward",
                "%s_incorrect"%ttype :"incorrect",
                "timeout":"incorrect" }
            #Associate all trial type endings to the end_trial function defined by Sequence
            #setattr(self, "_end_%s"%ttype, self._end_trial)

    def _start_picktrial(self):
        self.set_state(self.next_trial)

    def _start_incorrect(self):
        self.set_state("penalty")



'''
Eye tracking calibration task for head-mounted display
'''
import numpy as np

from riglib.experiment import traits, Sequence
from riglib.stereo_opengl.openxr import WindowVR
from riglib.stereo_opengl.window import Window
from .target_graphics import *

import zmq, msgpack, time


class CalibrateHMD(WindowVR, Sequence):
    """
    Show one target after another and send target locations to pupil-labs
    """

    status = dict(
        wait = dict(start_trial="target", start_pause="pause"),
        target = dict(fixation="target_calibrate", start_pause="pause"),
        target_calibrate = dict(target_done="wait", start_pause="pause", end_state=True), # end_state is just for counting trials
        pause = dict(end_pause="wait", end_state=True),
    )

    # initial state
    state = "wait"

    sequence_generators = ['target_generator']

    fixation_time = traits.Float(1.0, desc="Time in seconds to display target before calibration")
    calibration_time = traits.Float(1.0, desc="Duration in seconds of each target calibration")
    startup_time = traits.Float(5.0, desc="Time to wait before starting first trial")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target = VirtualCircularTarget(target_radius=1, target_color=[1,1,1,1])
        for model in self.target.graphics_models:
            self.add_model(model)

        print("Connecting to Pupil Service/Capture...")
        ctx = zmq.Context()

        # create a zmq REQ socket to talk to Pupil Service/Capture
        self.req = ctx.socket(zmq.REQ)
        self.req.connect("tcp://128.95.215.191:50020") # to-do don't hard code

        # set start eye windows
        n = {"subject": "eye_process.should_start.0", "eye_id": 0, "args": {}} # delay?
        print(self.send_recv_notification(n))
        n = {"subject": "eye_process.should_start.1", "eye_id": 1, "args": {}}
        print(self.send_recv_notification(n))
        time.sleep(2)

        n = {"subject": "set_detection_mapping_mode", "mode": "3d"}
        print(self.send_recv_notification(n))

        # set calibration method to hmd calibration
        n = {"subject": "start_plugin", "name": "HMD3DChoreographyPlugin", "args": {}}
        print(self.send_recv_notification(n))

        # start caliration routine with params. This will make pupil start sampeling pupil data.
        # the eye-translations have to be in mm, these here are default values from Unity XR
        n = {
            "subject": "calibration.should_start",
            "translation_eye0": [30., 0.0, 0.0],
            "translation_eye1": [-30., 0.0, 0.0],
            "record": True,
        }
        print(self.send_recv_notification(n))
        self.ref_data = []

    # pupil-labs convenience functions
    def send_recv_notification(self, n):
        # REQ REP requirese lock step communication with multipart msg (topic,msgpack_encoded dict)
        self.req.send_string("notify.%s" % n["subject"], flags=zmq.SNDMORE)
        self.req.send(msgpack.dumps(n, use_bin_type=True))
        return self.req.recv_string()

    def get_pupil_timestamp(self):
        self.req.send_string("t")  # see Pupil Remote Plugin for details
        return float(self.req.recv_string())

    def init(self):
        self.trial_dtype = np.dtype([('trial', 'u4'), ('index', 'u4'), ('target', 'f8', (3,))])
        super().init()

    def _cycle(self):
        '''
        Calls any update functions necessary and redraws screen
        '''
        super()._cycle()

    #### TEST FUNCTIONS ####
    def _test_start_trial(self, ts):
        if self.target_idx == 0:
            return ts > self.startup_time
        return not self.pause

    def _test_start_pause(self, ts):
        return self.pause

    def _test_fixation(self, ts):
        return ts > self.fixation_time

    def _test_target_done(self, ts):
        return ts > self.calibration_time
    
    def _test_end_pause(self, ts):
        return not self.pause

    def _parse_next_trial(self):
        '''Check that the generator has the required data'''
        self.target_idx, self.target_location = self.next_trial
        self.target_location = np.array(self.target_location).astype(float)
        
        # Update the data sinks with trial information
        self.trial_record['trial'] = self.calc_trial_num()
        self.trial_record['index'] = self.target_idx
        self.trial_record['target'] =  self.target_location
        self.sinks.send("trials", self.trial_record)

    ### STATE FUNCTIONS ###
    def _start_wait(self):
        self.sync_event('TARGET_OFF')
        self.target.hide()
        super()._start_wait()

    def _start_target(self):
        self.sync_event('TARGET_ON', 0xd)
        self.target.move_to_position(self.target_location)
        self.target.show()

    def _while_target_calibrate(self):
        # get the current pupil time (pupil uses CLOCK_MONOTONIC with adjustable timebase).
        # You can set the pupil timebase to another clock and use that.
        
        if self.cycle_count % 3 == 0:
            t = self.get_pupil_timestamp()

            # in this mockup  the left and right screen marker positions are identical.
            datum0 = {"mm_pos": (self.target_location[0]*10, self.target_location[2]*10, self.target_location[1]*10), "timestamp": t}
            self.ref_data.append(datum0)

    def _end_target_calibrate(self):
        # Send ref data to Pupil Capture/Service:
        # This notification can be sent once at the end or multiple times.
        # During one calibraiton all new data will be appended.
        n = {
            "subject": "calibration.add_ref_data",
            "ref_data": self.ref_data,
            "record": True,
        }
        print(self.send_recv_notification(n))
        self.ref_data = []

    def _start_pause(self):
        self.sync_event('PAUSE_START')
        self.target.hide()

    def _end_pause(self):
        self.sync_event('PAUSE_END')

    def _start_None(self):
        # stop calibration
        # Pupil will correlate pupil and ref data based on timestamps,
        # compute the gaze mapping params, and start a new gaze mapper.
        n = {
            "subject": "calibration.should_stop",
            "record": True,
        }
        print(self.send_recv_notification(n))


    @classmethod
    def get_desc(cls, params, log_summary):
        duration = round(log_summary['runtime'] / 60, 1)
        return "{}/{} succesful trials in {} min".format(
            log_summary['n_success_trials'], log_summary['n_trials'], duration)

    @staticmethod
    def target_generator(size=10):
        '''
        Generates a sequence of targets

        Parameters
        ----------
        size : int, optional
            Distance between targets, by default 10

        Returns
        -------
        [15 x 1] array of tuples containing trial index and [1 x 3] target coordinates

        '''
        pos = [
            (0.0, 0.0, 600.0),
            (0.0, 0.0, 1000.0),
            (0.0, 0.0, 2000.0),
            (180.0, 0.0, 600.0),
            (240.0, 0.0, 1000.0),
            (420.0, 0.0, 2000.0),
            (55.62306, 195.383, 600.0),
            (74.16407, 260.5106, 1000.0),
            (129.7871, 455.8936, 2000.0),
            (-145.6231, 120.7533, 600.0),
            (-194.1641, 161.0044, 1000.0),
            (-339.7872, 281.7577, 2000.0),
            (-145.6231, -120.7533, 600.0),
            (-194.1641, -161.0044, 1000.0),
            (-339.7872, -281.7577, 2000.0),
            (55.62306, -195.383, 600.0),
            (74.16407, -260.5106, 1000.0),
            (129.7871, -455.8936, 2000.0),
        ]
        pos = np.array(pos)/300
        pos = pos[:,[0,2,1]] - np.array([0,3,0])
        for idx in range(len(pos)):
            yield idx, np.array(pos[idx])*size




'''
Functions to call appropriate constructor functions based on UI data and to link decoder objects in the database
'''
import os
import re
import tempfile
import xmlrpc.client
import pickle
import numpy as np

from celery import chain
from .celery import app

from riglib.bmi import extractor

@app.task
def cache_plx(plxfile):
    """
    Create cache for plexon file
    """
    from plexon import plexfile
    plexfile.openFile(plxfile.encode('utf-8'))

@app.task
def make_bmi(name, clsname, extractorname, entry, cells, channels, binlen, tslice, ssm, pos_key, kin_extractor, zscore):
    """
    Create a new Decoder object from training data and save a record to the database

    Parameters
    ----------
    name : string
        Name assigned to decoder object in the database
    clsname : string
        BMI algorithm name (passed to bmilist lookup table 'bmis')
    extractorname : string
        feature extractor algorithm name (passed to bmilist lookup table 'extractors')
    entry : models.TaskEntry
        Django record of training task
    cells : string
        Single string containing all the units to be in decoder, matching
        format in global regex 'cellname' (used only for spike extractors)
    channels : string
        Single string containing all the channels to be in decoder; must be a
        comma separated list of values with spaces (e.g., "1, 2, 3")
        (used only for, e.g., LFP extractors)
    binlen : float
        Time of spike history to consider
    tslice : slice
        Task time to use when training the decoder
    ssm : string
        TODO
    pos_key : string
        TODO
    """
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
    from . import models
    from .json_param import Parameters
    from .tasktrack import Track
    from config import bmiconfig as namelist

    cellname = re.compile(r'(\d{1,3})\s*(\w{1})')

    print("make bmi")

    extractor_cls = namelist.extractors[extractorname]
    print('Training with extractor class:', extractor_cls)

    if 'spike' in extractor_cls.feature_type:  # e.g., 'spike_counts'
        # look at "cells" argument (ignore "channels")

        cells = [ (int(c), ord(u) - 96) for c, u in cellname.findall(cells)]
        if cells == []:
            units = None  # use all units by default
            # Note: inside training functions (e.g., _train_KFDecoder_manual_control,
            #   _train_KFDecoder_visual_feedback, etc.), remember to check if units
            #   variable is None, and if so, set the units from the plx file:
            #       if units == None:
            #           units = np.array(plx.units).astype(np.int32)"
        else:
            unique_cells = []
            for c in cells:
                if c not in unique_cells:
                    unique_cells.append(c)

            units = np.array(unique_cells).astype(np.int32)
    elif ('lfp' in extractor_cls.feature_type) or ('ai_' in extractor_cls.feature_type):  # e.g., 'lfp_power'
        # look at "channels" argument (ignore "cells")
        channels = np.array(channels.split(', ')).astype(np.int32)  # convert str to list of numbers
        if len(channels) == 0:
            channels = [1, 2, 3, 4]  # use these channels by default
        else:
            channels = np.unique(channels)
            #  units = np.hstack([channels.reshape(-1, 1), np.zeros(channels.reshape(-1, 1).shape, dtype=np.int32)])
            units = np.hstack([channels.reshape(-1, 1), np.ones(channels.reshape(-1, 1).shape, dtype=np.int32)])
    elif ('obs' in extractor_cls.feature_type):
        # ignore units and channels
        units = None
    else:
        raise Exception('Unknown extractor class!')

    # task_update_rate = 60 # NOTE may not be true for all tasks?!
    entry_data = models.TaskEntry.objects.get(id=entry).to_json()
    if hasattr(entry_data, 'params') and hasattr(entry_data['params'], 'fps'):
        task_update_rate = entry_data['params']['fps']
    else:
        task_update_rate = 60.

    extractor_kwargs = dict()
    if extractor_cls == extractor.BinnedSpikeCountsExtractor:
        extractor_kwargs['units'] = units
        extractor_kwargs['n_subbins'] = max(1, int((1./task_update_rate)/binlen))
    elif extractor_cls == extractor.LFPButterBPFPowerExtractor:
        extractor_kwargs['channels'] = channels
    elif extractor_cls == extractor.LFPMTMPowerExtractor:
        extractor_kwargs['channels'] = channels
    elif extractor_cls == extractor.AIMTMPowerExtractor:
        extractor_kwargs['channels'] = channels
    elif extractor_cls == extractor.DirectObsExtractor:
        pass
    else:
        raise Exception("Unknown extractor_cls: %s" % extractor_cls)

    # list of DataFile objects
    datafiles = models.DataFile.objects.filter(entry_id=entry)

    # key: a string representing a system name (e.g., 'plexon', 'blackrock', 'task', 'hdf')
    # value: a single filename, or a list of filenames if there are more than one for that system
    files = dict()
    system_names = set(d.system.name for d in datafiles)
    for system_name in system_names:
        filenames = [d.get_path() for d in datafiles if d.system.name == system_name]
        if system_name in ['blackrock', 'blackrock2']:
            files[system_name] = filenames  # list of (one or more) files
        else:
            assert(len(filenames) == 1)
            files[system_name] = filenames[0]  # just one file

    training_method = namelist.bmi_algorithms[clsname]
    ssm = namelist.bmi_state_space_models[ssm]
    kin_extractor_fn = namelist.kin_extractors[kin_extractor]
    decoder = training_method(files, extractor_cls, extractor_kwargs, kin_extractor_fn, ssm, units, update_rate=binlen, tslice=tslice, pos_key=pos_key,
        zscore=zscore, update_rate_hz=task_update_rate)
    decoder.te_id = entry

    tf = tempfile.NamedTemporaryFile('wb')
    pickle.dump(decoder, tf, 2)
    tf.flush()
    database = xmlrpc.client.ServerProxy("http://localhost:8000/RPC2/", allow_none=True)
    database.save_bmi(name, int(entry), tf.name)

def cache_and_train(*args, **kwargs):
    """
    Cache plexon file (if using plexon system) and train BMI.
    """
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
    from . import models

    recording_sys = models.KeyValueStore.get('recording_sys', None)
    if recording_sys == 'plexon':
        print("cache and train")
        entry = kwargs['entry']
        print(entry)
        plxfile = models.DataFile.objects.get(system__name='plexon', entry=entry)
        print(plxfile)

        if not plxfile.has_cache():
            cache = cache_plx.si(plxfile.get_path())
            train = make_bmi.si(*args, **kwargs)
            chain(cache, train)()
        else:
            print("calling")
            make_bmi.delay(*args, **kwargs)
    else:
        make_bmi.delay(*args, **kwargs)

def save_new_decoder_from_existing(obj, orig_decoder_record, suffix='_', dbname='default'):
    '''
    Save a decoder that is created by manipulating the parameters of an older decoder

    Parameters
    ----------
    obj: riglib.bmi.Decoder instance
        New decoder object to be saved
    orig_decoder_record: tracker.models.Decoder instance
        Database record of the original decoder
    suffix: string, default='_'
        The name of the new decoder is created by taking the name of the old decoder and adding the specified suffix

    Returns
    -------
    None
    '''
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
    from . import dbq
    import riglib.bmi
    if not isinstance(obj, riglib.bmi.bmi.Decoder):
        raise ValueError("This function is only intended for saving Decoder objects!")

    new_decoder_fname = obj.save()
    new_decoder_name = orig_decoder_record.name + suffix
    training_block_id = orig_decoder_record.entry_id
    print("Saving new decoder:", new_decoder_name)
    dbq.save_bmi(new_decoder_name, training_block_id, new_decoder_fname, dbname=dbname)

## Functions to manipulate existing (KF)Decoders. These belong elsewhere

def conv_mm_dec_to_cm(decoder_record):
    '''
    Convert a mm unit decoder to cm
    '''
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
    from . import dbq
    decoder_fname = os.path.join('/storage/decoders/', decoder_record.path)
    print(decoder_fname)
    decoder_name = decoder_record.name
    dec = pickle.load(open(decoder_fname))
    from riglib.bmi import train
    dec_cm = train.rescale_KFDecoder_units(dec, 10)

    new_decoder_basename = os.path.basename(decoder_fname).rstrip('.pkl') + '_cm.pkl'
    new_decoder_fname = '/tmp/%s' % new_decoder_basename
    pickle.dump(dec_cm, open(new_decoder_fname, 'w'))

    new_decoder_name = decoder_name + '_cm'
    training_block_id = decoder_record.entry_id
    print(new_decoder_name)
    dbq.save_bmi(new_decoder_name, training_block_id, new_decoder_fname)

def zero_out_SSKF_bias(decoder_record):
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'

    dec = open_decoder_from_record(decoder_record)
    dec.filt.C_xpose_Q_inv_C[:,-1] = 0
    dec.filt.C_xpose_Q_inv_C[-1,:] = 0
    save_new_decoder_from_existing(dec, decoder_record, suffix='_zero_bias')

def conv_kfdecoder_binlen(decoder_record, new_binlen):
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
    dec = open_decoder_from_record(decoder_record)
    dec.change_binlen(new_binlen)
    save_new_decoder_from_existing(dec, decoder_record, suffix='_%dHz' % int(1./new_binlen))

def conv_kfdecoder_to_ppfdecoder(decoder_record):
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
    from . import dbq

    # Load the decoder
    decoder_fname = os.path.join('/storage/decoders/', decoder_record.path)
    print(decoder_fname)
    decoder_name = decoder_record.name
    dec = pickle.load(open(decoder_fname))

    from riglib.bmi import train
    dec_ppf = train.convert_KFDecoder_to_PPFDecoder(dec)

    new_decoder_basename = os.path.basename(decoder_fname).rstrip('.pkl') + '_ppf.pkl'
    new_decoder_fname = '/tmp/%s' % new_decoder_basename
    pickle.dump(dec_ppf, open(new_decoder_fname, 'w'))

    new_decoder_name = decoder_name + '_ppf'
    training_block_id = decoder_record.entry_id
    print(new_decoder_name)
    from . import dbq
    dbq.save_bmi(new_decoder_name, training_block_id, new_decoder_fname)

def conv_kfdecoder_to_sskfdecoder(decoder_record):
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'

    dec = open_decoder_from_record(decoder_record)

    F, K = dec.filt.get_sskf()
    from riglib.bmi import sskfdecoder
    filt = sskfdecoder.SteadyStateKalmanFilter(F=F, K=K)
    dec_sskf = sskfdecoder.SSKFDecoder(filt, dec.units, dec.ssm, binlen=decoder.binlen)

    save_new_decoder_from_existing(decoder_record, '_sskf')

def make_kfdecoder_interpolate(decoder_record):
    os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
    from . import dbq

    # Load the decoder
    decoder_fname = os.path.join('/storage/decoders/', decoder_record.path)
    print(decoder_fname)
    decoder_name = decoder_record.name
    dec = pickle.load(open(decoder_fname))

    from riglib.bmi import train
    dec_ppf = train._interpolate_KFDecoder_state_between_updates(dec)

    new_decoder_basename = os.path.basename(decoder_fname).rstrip('.pkl') + '_ppf.pkl'
    new_decoder_fname = '/tmp/%s' % new_decoder_basename
    pickle.dump(dec_ppf, open(new_decoder_fname, 'w'))

    new_decoder_name = decoder_name + '_60hz'
    training_block_id = decoder_record.entry_id
    print(new_decoder_name)
    from . import dbq
    dbq.save_bmi(new_decoder_name, training_block_id, new_decoder_fname)




'''Needs docs'''


import numpy as np
from OpenGL.GL import *

from ..textures import Texture

_mattypes = {
    (4,4):"4", (3,3):"3", (2,2):"2", 
    (2,3):"2x3",(3,2):"3x2",
    (2,4):"2x4",(4,2):"4x2",
    (3,4):"3x4",(4,3):"4x3"
}
_typename = { int:"i", float:"f" }
class _getter(object):
    '''Wrapper object which allows direct getting and setting of shader values'''
    def __init__(self, type, prog):
        setter = super(_getter, self).__setattr__
        
        setter("prog", prog)
        setter("cache", dict())
        setter("func", globals()['glGet{type}Location'.format(type=type)])

        if type == "Uniform":
            setter("info", dict())
            for i in range(glGetProgramiv(self.prog, GL_ACTIVE_UNIFORMS)):
                name, size, t = glGetActiveUniform(self.prog, i)
                self.info[name] = t
    
    def __getattr__(self, attr):
        if attr not in self.cache:
            self.cache[attr] = self.func(self.prog, attr)
        return self.cache[attr]
    
    def __getitem__(self, attr):
        if attr not in self.cache:
            self.cache[attr] = self.func(self.prog, attr)
        return self.cache[attr]

    def __setitem__(self, attr, val):
        self._set(attr, val)
    
    def __setattr__(self, attr, val):
        self._set(attr, val)
    
    def __contains__(self, attr):
        if attr not in self.cache:
            self.cache[attr] = self.func(self.prog, attr)
        return self.cache[attr] != -1
    
    def _set(self, attr, val):
        '''This heinously complicated function has to guess the function to use because
        there are no strong types in python, hence we just have to guess'''
        if attr not in self.cache:
            self.cache[attr] = self.func(self.prog, attr)

        if isinstance(val, Texture):
            # Handle texture uniforms
            glUniform1i(self.cache[attr], val.texunit)

        elif isinstance(val, np.ndarray) and len(val.shape) > 1:
            assert len(val.shape) <= 3
            if val.shape[-2:] in _mattypes:
                nmats = val.shape[0] if len(val.shape) == 3 else 1
                fname = _mattypes[val.shape[-2:]]
                func = globals()['glUniformMatrix%sfv'%fname]
                #We need to transpose all numpy matrices since numpy is row-major
                #and opengl is column-major
                func(self.cache[attr], nmats, GL_TRUE, val.astype(np.float32).ravel())
        elif isinstance(val, (list, tuple, np.ndarray)):
            #glUniform\d[if]v
            if isinstance(val[0], (tuple, list, np.ndarray)):
                assert len(val[0]) <= 4
                t = _typename[type(val[0][0])]
                func = globals()['glUniform%d%sv'%(len(val[0]), t)]
                func(self.cache[attr], len(val), np.array(val).astype(np.float32).ravel())
            else:
                t = _typename[type(val[0])]
                func = globals()['glUniform%d%s'%(len(val), t)]
                func(self.cache[attr], *val)
        elif isinstance(val, (int, float)):
            # Handle scalar uniforms
            try:
                if attr in ['nearclip', 'farclip']:  # Force these to be floats
                    glUniform1f(self.cache[attr], float(val))
                else:
                    t = _typename[type(val)]
                    globals()[f'glUniform1{t}'](self.cache[attr], val)
            except GLError as e:
                print(f"OpenGL Error when setting uniform '{attr}': {e}")
                print(f"Uniform location: {self.cache[attr]}, Value: {val}, Type: {type(val)}")
        else:
            print(f"Warning: Unsupported type {type(val)} for uniform '{attr}'")

class ShaderProgram(object):
    def __init__(self, shaders):
        self.shaders = shaders
        self.program = glCreateProgram()
        for shader in shaders:
            glAttachShader(self.program, shader)
        glLinkProgram(self.program)

        if not glGetProgramiv(self.program, GL_LINK_STATUS):
            err = glGetProgramInfoLog(self.program)
            glDeleteProgram(self.program)
            raise Exception(err)
        
        self.attributes = _getter("Attrib", self.program)
        self.uniforms = _getter("Uniform", self.program)
    
    def draw(self, ctx, models, **kwargs):
        glUseProgram(self.program)
        try:
            for name, v in list(kwargs.items()):
                if isinstance(v, Texture):
                    self.uniforms[name] = ctx.get_texunit(v)
                elif name in self.uniforms:
                    self.uniforms[name] = v
                elif name in self.attributes:
                    self.attributes[name] = v
                elif hasattr(v, "__call__"):
                    v(self)
            
            for tex, funcs in list(models.items()):
                if tex is None:
                    self.uniforms.textureSampler = ctx.get_texunit("None")
                else:
                    self.uniforms.textureSampler = ctx.get_texunit(tex)

                for drawfunc in funcs:
                    drawfunc(self)
        finally:
            glUseProgram(0)  # Ensure we always unbind the program


'''
Web browser GUI-launched tasks run in a separate process. This module provides
mechanisms for interacting withe the task running in another process, e.g.,
calling functions to start/stop the task, enabling/disabling decoder adaptation, etc.
'''
import os
import sys
import io
import traceback
from datetime import datetime
import multiprocessing as mp

from . import websocket
from riglib import experiment, singleton
from .json_param import Parameters

log_path = os.path.join(os.path.dirname(__file__), '../../log')
log_filename = os.path.join(log_path, "tasktrack_log")
# TODO make folder if it doesn't exist

def log_error(err, mode='a'):
    traceback.print_exc(None, err)
    with open(log_filename, mode) as fp:
        err.seek(0)
        fp.write(err.read())

def log_str(s, mode="a", newline=True):
    if newline and not s.endswith("\n"):
        s += "\n"
    with open(log_filename, mode) as fp:
        fp.write(s)


if sys.platform == "win32":
    use_websock = False
else:
    use_websock = True

class Track(singleton.Singleton):
    '''
    Tracker for task instantiation running in a separate process. This is a singleton.
    '''
    __instance = None
    def __init__(self, use_websock=use_websock):
        super().__init__()
        # shared memory to store the status of the task in a char array
        self.status = mp.Array('c', 256)
        self.task_proxy = None
        self.proc = None
        self.use_websock = use_websock
        self.websock = None

    def start_websock(self):
        if self.websock is None and self.use_websock:
            print("Starting websocket...")
            self.websock = websocket.Server(self.notify)

    def notify(self, msg):
        if msg['status'] == "error" or msg['State'] == "stopped":
            self.status.value = b""

    def runtask(self, base_class=experiment.Experiment, feats=[], cli=False, **kwargs):
        '''
        Begin running of task
        '''
        now = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        log_str("\n{}\n-------------------\nRunning new task: {}\n".format(now, base_class))

        if not cli:
            self.start_websock()

        if None in feats:
            raise ValueError("Features not found properly in database!")

        # initialize task status
        self.status.value = b"running" if 'saveid' in kwargs else b"testing"

        if 'seq' in kwargs:
            kwargs['seq_params'] = kwargs['seq'].params
            kwargs['seq'] = kwargs['seq'].get()  ## retreive the database data on this end of the pipe

        use_websock = not (self.websock is None)
        if use_websock:
            feats.insert(0, websocket.NotifyFeat)
            kwargs['params']['websock'] = self.websock
            kwargs['params']['tracker_status'] = self.status

        task_class = experiment.make(base_class, feats=feats)

        # process parameters
        params = kwargs['params']
        if isinstance(params, str):
            params = Parameters(params)
        elif isinstance(params, dict):
            params = Parameters.from_dict(params)
        params.trait_norm(task_class.class_traits())
        params = params.params # dict
        kwargs.pop('params', None)

        log_str("Spawning process...")
        log_str(str(kwargs))

        # Spawn the process
        self.proc = experiment.task_wrapper.TaskWrapper(
            log_filename=log_filename, params=params,
            target_class=task_class, websock=self.websock, status=self.status,
            **kwargs)

        # create a proxy for interacting with attributes/functions of the task.
        # The task runs in a separate process and we cannot directly access python
        # attributes of objects in other processes
        self.task_proxy, _ = self.proc.start()

    def __del__(self):
        '''
        Destructor for Track object. Not sure if this function ever gets called
        since Track is a singleton created upon import of the db.tracker.ajax module...
        '''
        if self.websock is not None:
            self.websock.stop()

    def stoptask(self):
        '''
        Terminate the task gracefully by running riglib.experiment.Experiment.end_task
        '''
        self.update_alive()            
        if self.status.value == b"":
            print("Task already stopped!")
            return
        elif self.status.value not in [b"testing", b"running"]:
            raise ValueError("Unknown task status")

        assert self.status.value in [b"testing", b"running"]
        print("Tasktrack is stopping the task...")
        try:
            if self.task_proxy is not None:
                self.task_proxy.end_task()
        except Exception as e:
            print(e)
            traceback.print_exc()
            err = io.StringIO()
            traceback.print_exc(None, err)
            err.seek(0)
            return dict(status="error", msg=err.read())

        status = self.status.value.decode("utf-8")
        self.reset()

        '''
        WIP cloud upload code
        # allow time for files, etc. to be saved
        time.sleep(3)

        if self.proc.saveid is not None:
            from db.tracker import models
            te = models.TaskEntry.objects.get(id=self.proc.saveid)

            # Wrap up HDF file saving
            models.DataFile.objects.get(entry__id=te.id, system__name="hdf")
            metadata = dict(
                task_name = te.task.name,
                block_number = te.id,
            )

            # upload metadata to server, if appropriate
            print("Attempting to save to cloud...")
            te.upload_to_cloud()
        '''

        return status

    def reset(self):
        self.task_proxy = None
        self.status.value = b""

    def get_status(self):
        return self.status.value.decode("utf-8")

    def update_alive(self):
        """ Check if the remote process is still alive, and if dead, reset the task_proxy object """
        if (not self.proc is None) and (not self.proc.is_alive()):
            print("process died in error, destroying proxy object")
            self.reset()

    def task_running(self):
        print(self.get_status())
        return self.get_status() in ["running", "testing"]


import numpy as np
import time
from riglib.gpio import ArduinoGPIO, DigitalWave, TeensyGPIO, convert_masked_data_to_pins
from riglib import source
from riglib.ecube import Digital, LFP
from features.sync_features import ArduinoSync
from config.rig_defaults import rig2_sync_params_arduino, human_sync_params_arduino
from config.rig_defaults import reward
import aopy

import unittest

sync_params = human_sync_params_arduino # rig2_sync_params_arduino # rig1_sync_params_arduino

class TestDIO(unittest.TestCase):

    @unittest.skip("")
    def test_reward_out(self):
        dio = ArduinoGPIO('/dev/teensydio')
        time.sleep(5)
        
        for pin in range(2,14):
            print(pin)
            # Send a bunch of pulses
            pulse_width = 0.2
            pulse_interval = 0.2
            duration = 1
            t0 = time.perf_counter()
            while time.perf_counter() - t0 < duration:
                dio.write(pin, 1)
                t1 = time.perf_counter()
                while time.perf_counter() - t1 < pulse_width:
                    time.sleep(0.001)
                dio.write(pin, 0)
                t1 = time.perf_counter()
                while time.perf_counter() - t1 < pulse_interval:
                    time.sleep(0.001)

    def test_dio(self):
        print("Testing direct connection to DIO")

        # Connect to the DIO
        mask = sync_params['event_sync_mask']
        print(f'using mask: {mask}')
        shift = sync_params['event_sync_data_shift']
        dio = TeensyGPIO('/dev/teensydio', baudrate=57600)
        dio.write_many(mask, 0)

        pulse_data = mask
        pins, values = convert_masked_data_to_pins(mask, pulse_data)
        print(pins, values)

        time.sleep(5)
        print("Pulsing")
        
        # Send a bunch of pulses
        pulse_width = 0.2
        pulse_interval = 0.2
        duration = 5
        t0 = time.perf_counter()
        while time.perf_counter() - t0 < duration:
            dio.write_many(mask, pulse_data)
            t1 = time.perf_counter()
            while time.perf_counter() - t1 < pulse_width:
                time.sleep(0.001)
            dio.write_many(mask, 0)
            t1 = time.perf_counter()
            while time.perf_counter() - t1 < pulse_interval:
                time.sleep(0.001)

    @unittest.skip("")
    def test_dio_speed_direct(self):
        print("Testing direct connection to DIO")

        # Connect to the DIO
        mask = sync_params['event_sync_mask']
        print(f'using mask: {mask}')
        shift = sync_params['event_sync_data_shift']
        dio = TeensyGPIO('/dev/ttyACM0', baudrate=115200)
        dio.write_many(mask, 0)

        # Record data from the digital sync channels
        channels = [ch+1 for ch in sync_params['event_sync_dch']]
        ds = source.MultiChanDataSource(Digital, channels=channels, bufferlen=10)
        ds.start()
        time.sleep(1)
        
        # Send a bunch of pulses
        pulse_width = 0.003
        pulse_data = 255
        pulse_interval = 0.003
        duration = 5
        t0 = time.perf_counter()
        while time.perf_counter() - t0 < duration:
            dio.write_many(mask, pulse_data << shift)
            t1 = time.perf_counter()
            while time.perf_counter() - t1 < pulse_width:
                time.sleep(0)
            dio.write_many(mask, 0)
            t1 = time.perf_counter()
            while time.perf_counter() - t1 < pulse_interval:
                time.sleep(0)

        # Record the data
        time.sleep(1)
        data = ds.get_new(channels)
        ds.stop()
        ecube_dig_channels = np.squeeze(data).T
        nch = len(channels)
        digital_samplerate = 25000

        # Make sure all the pulses have the right data
        digital_data = np.squeeze(np.packbits(ecube_dig_channels, bitorder='little').view(np.dtype('<u8')))
        # event_bit_mask = aopy.utils.convert_channels_to_mask(channels)
        ecube_sync_data = aopy.utils.mask_and_shift(digital_data, 0xff)
        raw_times, raw_events = aopy.utils.detect_edges(ecube_sync_data, digital_samplerate, rising=True, falling=False)
        print(f"Recorded {len(raw_times)} events\n---------------------------")
        np.testing.assert_allclose(raw_events, pulse_data)

        # Separate individual pulses into "ok", "too_short" and "too_long"
        pulses_ok = []
        pulses_too_short = []
        pulses_too_long = []

        for ch in range(nch):
            raw_times, raw_events = aopy.utils.detect_edges(ecube_dig_channels[:,ch], digital_samplerate, rising=True, falling=True)
            
            widths = raw_times[1::2]-raw_times[::2]
            
            width_too_short = widths < 0.002
            width_too_long = widths > 0.004
            width_ok = (~width_too_short) & (~width_too_long)
            
            # Save the timestamps of the events
            pulses_too_short.append(raw_times[::2][width_too_short])
            pulses_too_long.append(raw_times[::2][width_too_long])
            pulses_ok.append(raw_times[::2][width_ok])

        # Collect all the channels together
        width_too_short = np.concatenate(pulses_too_short, axis=0)
        print(f"too short: {np.count_nonzero(width_too_short)} pulses")
        width_too_long = np.concatenate(pulses_too_long, axis=0)
        print(f"too long: {np.count_nonzero(width_too_long)} pulses")
        width_ok = np.concatenate(pulses_ok, axis=0)
        print(f"ok: {np.count_nonzero(width_ok)} pulses")

    @unittest.skip("")
    def test_dio_speed_multithreaded(self):
        print("Testing DIO via ArduinoSync()")

        # Use the sync feature to connect to the DIO
        dio = ArduinoSync()

        # Record data from the digital sync channels
        channels = dio.sync_params['event_sync_dch']
        ds = source.MultiChanDataSource(Digital, channels=channels, bufferlen=10)
        ds.start()
        time.sleep(1)
        
        # Send a bunch of pulses
        pulse_width = 0.003
        pulse_data = 255
        pulse_interval = 0.003
        duration = 5
        t0 = time.perf_counter()
        while time.perf_counter() - t0 < duration:
            dio.sync_code(int(pulse_data) << dio.sync_params['event_sync_data_shift'])
            t1 = time.perf_counter()
            while time.perf_counter() - t1 < pulse_width + pulse_interval:
                time.sleep(0)

        # Record the data
        time.sleep(1)
        data = ds.get_new(channels)
        ds.stop()
        ecube_dig_channels = np.array(data).T
        nch = len(channels)
        digital_samplerate = 25000

        # Make sure all the pulses have the right data
        digital_data = np.squeeze(np.packbits(ecube_dig_channels, bitorder='little').view(np.dtype('<u8')))
        event_bit_mask = aopy.utils.convert_channels_to_mask(channels)
        ecube_sync_data = aopy.utils.mask_and_shift(digital_data, event_bit_mask)
        raw_times, raw_events = aopy.utils.detect_edges(ecube_sync_data, digital_samplerate, rising=True, falling=False)
        print(f"Recorded {len(raw_times)} events\n---------------------------")
        np.testing.assert_allclose(raw_events, pulse_data)

        # Separate individual pulses into "ok", "too_short" and "too_long"
        pulses_ok = []
        pulses_too_short = []
        pulses_too_long = []

        for ch in range(nch):
            raw_times, raw_events = aopy.utils.detect_edges(ecube_dig_channels[:,ch], digital_samplerate, rising=True, falling=True)
            
            widths = raw_times[1::2]-raw_times[::2]
            
            width_too_short = widths < 0.002
            width_too_long = widths > 0.004
            width_ok = (~width_too_short) & (~width_too_long)
            
            # Save the timestamps of the events
            pulses_too_short.append(raw_times[::2][width_too_short])
            pulses_too_long.append(raw_times[::2][width_too_long])
            pulses_ok.append(raw_times[::2][width_ok])

        # Collect all the channels together
        width_too_short = np.concatenate(pulses_too_short, axis=0)
        print(f"too short: {np.count_nonzero(width_too_short)} pulses")
        width_too_long = np.concatenate(pulses_too_long, axis=0)
        print(f"too long: {np.count_nonzero(width_too_long)} pulses")
        width_ok = np.concatenate(pulses_ok, axis=0)
        print(f"ok: {np.count_nonzero(width_ok)} pulses")

if __name__ == '__main__':
    unittest.main()


import socket
import re
import numpy as np
from riglib.oculomatic import System
import time

UDP_IP = ""
UDP_PORT = 11999

# Format is "frame_count&le_x&le_y&le_diam&re_x&re_y&re_diam"

def test_streaming_direct():
    sock = socket.socket(socket.AF_INET, # Internet
                        socket.SOCK_DGRAM) # UDP
    sock.bind((UDP_IP, UDP_PORT))

    while True:
        data, addr = sock.recvfrom(1024) # buffer size is 1024 bytes
        items = data.decode('utf-8').split('&')
        frame_count = int(items[0])
        try:
            le_x = float(items[1])
        except:
            le_x = np.nan
        try:
            le_y = float(items[2])
        except:
            le_y = np.nan
        le_diam = int(items[3])
        try:
            re_x = float(items[4])
        except:
            re_x = np.nan
        try:
            re_y = float(items[5])
        except:
            re_y = np.nan
        re_diam = int(items[6])

        print(f"frame {frame_count:0>8d}: left eye ({le_x:+06.2f}, {le_y:+06.2f}) {le_diam:0>5d}, right eye ({re_x:+06.2f}, {re_y:+06.2f}) {re_diam:0>5d}")

        
def test_streaming_class():
    from riglib import source
    eyedata = source.DataSource(System)
    eyedata.start()
    time.sleep(0.5)

    data = eyedata.get()
    eyedata.stop()

    print(data)

test_streaming_class()





'''Needs docs'''
import numpy as np
from .textures import Texture

def frustum(l, r, t, b, n, f):
    '''
    This function emulates glFrustum: https://www.opengl.org/sdk/docs/man2/xhtml/glFrustum.xml
    A frustum is a solid cut by planes, e.g., the planes representing the viewable area of a screen.

    Parameters
    ----------
    l: float
        Distance to the left plane of the screen
    r: float
        Distance to the right plane of the screen
    t: float
        Distance to the top plane of the screen
    b: float
        Distance to the bottom plane of the screen
    n: float
        Distance to the near plane of the screen
    f: float
        Distance to the far plane of the screen

    Returns
    -------
    Projection matrix to apply to solid to truncate

    '''
    rl, nrl = r + l, r - l
    tb, ntb = t + b, t - b
    fn, nfn = f + n, f - n
    return np.array([[2*n / nrl,     0,     rl / nrl,       0],
                     [    0,     2*n / ntb, tb / ntb,       0],
                     [    0,         0,    -fn / nfn, -2*f*n / nfn],
                     [    0,         0,       -1,           0]])

def perspective(angle, aspect, near, far):
    '''Generates a perspective transform matrix'''
    f = 1./ np.tan(np.radians(angle) / 2)
    fn = far + near
    nfn = far - near
    return np.array([[f/aspect, 0,    0,               0],
                     [0,        f,    0,               0],
                     [0,        0, -fn/nfn, -2*far*near/nfn],
                     [0,        0,   -1,               0]])

def orthographic(w, h, near, far):
    fn = far + near
    nfn = far - near
    return np.array([[2/w, 0,   0,      0],
                     [0,   2/h, 0,      0],
                     [0,   0,   -2/nfn, -fn/nfn],
                     [0,   0,   0,       1]])

def offaxis_frusta(winsize, fov, near, far, focal_dist, iod, flip=False, flip_z=False):
    aspect = winsize[0] / winsize[1]
    top = near * np.tan(np.radians(fov) / 2)
    right = aspect*top
    fshift = (iod/2) * near / focal_dist

    # calculate the perspective matrix for the left eye and for the right eye
    left = frustum(-right+fshift, right+fshift, top, -top, near, far)
    right = frustum(-right-fshift, right-fshift, top, -top, near, far)
    
    # multiply in the iod (intraocular distance) modelview transform
    lxfm, rxfm = np.eye(4), np.eye(4)
    lxfm[:3,-1] = [0.5*iod, 0, 0]
    rxfm[:3,-1] = [-0.5*iod, 0, 0]
    flip_mat = np.eye(4)


    if flip:
        flip_mat[0,0] = -1
    if flip_z:
        flip_mat[1,1] = -1

    return np.dot(flip_mat, np.dot(left, lxfm)), np.dot(flip_mat, np.dot(right, rxfm))

    #return np.dot(left, lxfm), np.dot(right, rxfm)

def cloudy_tex(size=(512,512), alpha=0.5):
    '''Generates 1/f distributed noise and puts it into a texture. Looks like clouds'''
    im = np.random.randn(*size)
    grid = np.mgrid[-1:1:size[0]*1j, -1:1:size[1]*1j]
    mask = 1/(grid**2).sum(0)
    fim = np.fft.fftshift(np.fft.fft2(im))
    im = np.abs(np.fft.ifft2(np.fft.fftshift(mask * fim)))
    im -= im.min()
    im /= im.max()
    im = np.tile(im, (4,1,1)).T
    im[:,:,3] = alpha
    return Texture(im)

def create_grid_texture(size=800, density=50, thickness=3, line_color=[0.5, 0.5, 0.5, 1], 
                        background_color=[0.1, 0.1, 0.1, 1]):
    """
    Create a grid texture.
    """
    size = int(size)
    thickness = int(thickness)
    grid_texture = np.ones((size, size, 4)) * background_color  # RGB image
    num_lines = (size + 1)//density
    line_spacing = size//num_lines

    # Draw horizontal grid lines
    for r in range(num_lines):
        start = int(r * line_spacing - thickness//2)
        grid_texture[start:start+thickness, :, :] = line_color
        grid_texture[:, start:start+thickness, :] = line_color

    return Texture(grid_texture/np.max(grid_texture), mipmap=True, anisotropic_filtering=2)

def look_at(eye, target, up):
    # Convert inputs to numpy arrays
    eye = np.array(eye)
    target = np.array(target)
    up = np.array(up)

    # Calculate forward direction
    forward = target - eye
    forward = forward / np.linalg.norm(forward)

    # Calculate right direction
    right = np.cross(forward, up)
    right = right / np.linalg.norm(right)

    # Recalculate up direction
    up = np.cross(right, forward)

    # Create rotation matrix
    rotation = np.array([
        [right[0], right[1], right[2], 0],
        [up[0], up[1], up[2], 0],
        [-forward[0], -forward[1], -forward[2], 0],
        [0, 0, 0, 1]
    ])

    # Create translation matrix
    translation = np.array([
        [1, 0, 0, -eye[0]],
        [0, 1, 0, -eye[1]],
        [0, 0, 1, -eye[2]],
        [0, 0, 0, 1]
    ])

    # Combine rotation and translation
    return np.dot(rotation, translation)


"""
Modified by Pavi for Orsborn ab reward system
on Aug 2020 based on Si Jia's previous script

"""
from riglib import reward

import unittest

# class TestAoReward(unittest.TestCase):

#     def setUp(self):
#         self.reward_sys = reward.Basic()

#     def test_connection(self):
#         self.assertTrue(self.reward_sys.board is not None)

#     def test_calibrate(self):
#         import numpy as np
#         import time

#         n_reps = 10
#         for t in np.arange(0.1, 1.1, 0.1):
#             print("Testing " + str(t) + " seconds")
#             for i in range(n_reps):
#                 print(".", end="")
#                 self.reward_sys.drain(t)
#                 time.sleep(0.5)


class TestAoPellet(unittest.TestCase):

    def setUp(self):
        self.reward_sys = reward.pellet_open()

    def test_connection(self):
        self.assertTrue(self.reward_sys is not None)
        self.assertTrue(self.reward_sys.board is not None)

    def test_calibrate(self):
        import numpy as np
        import time

        n_reps = 10
        for i in range (n_reps):
            print(".", end="")
            self.reward_sys.dispense()
            time.sleep(1)


if __name__ == '__main__':
    unittest.main()






import cProfile
import pstats
import socket
import traceback
from riglib.experiment import traits
import numpy as np
import json

class Profiler():
    
    def run(self):
        pr = cProfile.Profile()
        pr.enable()
        super().run()
        pr.disable()
        with open('profile.csv', 'w') as f:
            ps = pstats.Stats(pr, stream=f).sort_stats('time')
            ps.print_stats()

class OnlineAnalysis(traits.HasTraits):
    '''
    Feature to send task data to an online analysis server.

    In the future this could be expanded to make use of the riglib.sinks interface.
    For now it is a simple UDP socket interface that sends messages about the 
    task params, state transitions, and sync events to a server for online analysis.
    '''

    online_analysis_ip = traits.String("localhost", desc="IP address of the machine running the online analysis")
    online_analysis_port = traits.Int(5000, desc="Port number for the online analysis server")
        
    def _send_online_analysis_msg(self, key, *values):
        '''
        Helper function to send messages to the online analysis server
        '''
        strings = []
        for v in values:
            if isinstance(v, np.ndarray) and v.size > 1:
                strings.append(json.dumps(v.tolist()))
            elif isinstance(v, np.ndarray):
                strings.append(json.dumps(v[0]))
            elif isinstance(v, np.generic):
                strings.append(json.dumps(v.item()))
            else:
                try:
                    strings.append(json.dumps(v))
                except (TypeError, OverflowError):
                    print('Could not convert to json:', v, 'for key:', key)
                    traceback.print_exc()
                    strings.append('null')
        payload = '#'.join(strings)
        self.online_analysis_sock.sendto(f'{key}%{payload}'.encode('utf-8'), (self.online_analysis_ip, self.online_analysis_port))

    def init(self):
        '''
        Send basic experiment info to the online analysis server
        '''
        super().init()
        try:
            self.online_analysis_sock = socket.socket(
                socket.AF_INET, # Internet
                socket.SOCK_DGRAM) # UDP
            self.online_analysis_sock.setblocking
            self._send_online_analysis_msg('init', False) # Just a test message
        except:
            print('Could not connect to socket')
            return
        
        # Send entry metadata
        try:
            self._send_online_analysis_msg('param', 'experiment_name', self.__class__.__name__)
            if hasattr(self, 'saveid'):
                self._send_online_analysis_msg('param', 'te_id', self.saveid)
            else:
                self._send_online_analysis_msg('param', 'te_id', 'None')
            if hasattr(self, 'subject_name'):
                self._send_online_analysis_msg('param', 'subject_name', self.subject_name)
        except:
            print('Problem sending entry metadata')
            traceback.print_exc()

        # Send task metadata
        for key, value in self.get_trait_values().items():
            try:
                if key in self.object_trait_names:
                    self._send_online_analysis_msg('param', key, None) # Skip objects
                    if key == 'decoder':
                        self._send_online_analysis_msg('param', 'decoder_channels', value.channels.flatten().tolist())
                        self._send_online_analysis_msg('param', 'decoder_states', value.states)
                        self._send_online_analysis_msg('param', 'decoder_bands', [(0,0)]) # TODO: How to get this?
                else:
                    self._send_online_analysis_msg('param', key, value)
            except:
                print('Problem sending task metadata')
                print(key, value)
                traceback.print_exc()

        if hasattr(self, 'sync_params'):
            for key, value in self.sync_params.items():
                try:
                    self._send_online_analysis_msg('param', key, value)
                except:
                    print('Problem sending sync param')
                    print(key, value)
                    traceback.print_exc()

        # Send init message to trigger analysis workers
        self._send_online_analysis_msg('init', True)
        print('...done!')

    def _start_wait(self):
        if hasattr(super(), '_start_wait'):
            super()._start_wait()
        if hasattr(self, 'targs') and hasattr(self, 'gen_indices'):
            for i in range(len(self.targs)):
                self._send_online_analysis_msg('target_location', self.gen_indices[i], self.targs[i])

    def _cycle(self):
        '''
        Send cursor and eye position data to the online analysis server
        '''
        super()._cycle()
        self._send_online_analysis_msg('cycle_count', self.cycle_count)
        if hasattr(self, 'plant'):
            self._send_online_analysis_msg('cursor', self.plant.get_endpoint_pos())
        if hasattr(self, 'eye_pos'):
            self._send_online_analysis_msg('eye_pos', self.eye_pos)
        if hasattr(self, 'task_data') and 'decoder_state' in self.task_data.dtype.names:
            self._send_online_analysis_msg('decoder_state', self.task_data['decoder_state'].flatten().tolist())
        if hasattr(self, 'task_data') and hasattr(self, 'extractor') and self.extractor.feature_type in self.task_data.dtype.names:
            self._send_online_analysis_msg('neural_features', self.task_data[self.extractor.feature_type].flatten().tolist())

    def set_state(self, condition, **kwargs):
        '''
        Send task state transitions to the online analysis server
        '''
        self._send_online_analysis_msg('state', condition)
        super().set_state(condition, **kwargs)

    def sync_event(self, event_name, event_data=0, immediate=False):
        '''
        Send sync events to the online analysis server
        '''
        if not immediate:
            self._send_online_analysis_msg('sync_event', event_name, event_data)
        super().sync_event(event_name, event_data=event_data, immediate=immediate)

class ReplayCursor():

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.replay_cursor_data = kwargs.pop('replay_cursor_data')

    def move_effector(self, *args, **kwargs):
        if self.cycle_count >= len(self.replay_cursor_data):
            self.state = None
            return
        pos = self.replay_cursor_data[self.cycle_count]
        self.task_data['manual_input'] = pos
        self.plant.set_endpoint_pos(pos)

class ReplayEye():

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.replay_eye_data = kwargs.pop('replay_eye_data')

    def _cycle(self):
        super()._cycle()
        if self.cycle_count >= len(self.replay_eye_data):
            self.state = None
            return
        pos = self.replay_eye_data[self.cycle_count]
        self.eye_pos = pos
        if 'eye' in self.task_data.dtype.names:
            self.task_data['eye'] = pos


# Generated by Django 3.1.4 on 2021-10-17 15:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0009_auto_20210630_1019'),
    ]

    operations = [
        migrations.CreateModel(
            name='Experimenter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
            ],
        ),
    ]



'''
Base code for 'optitrack' feature, compatible with Optitrack motiontracker
'''
import time
import numpy as np
from ..source import DataSourceSystem

class System(DataSourceSystem):
    '''
    Optitrack DataSourceSystem collects motion tracking data via UDP packets using natnet depacketizer
    '''
    update_freq = 240 # This may not always be the case, but lower frequencies will still work, just waste space in the circular buffer
    
    def __init__(self, client, feature="rigid body", n_features=1):
        '''
        Don't start the client in this DataSourceSystem object since then it won't be available for 
        commands elsewhere, i.e. start/stop recording
        '''
        self.client = client
        self.feature = feature # rigid body, skeleton, marker
        self.n_features = n_features
        self.rigid_bodies = []
        self.skeletons = []
        self.markers = []
        self.timing = []
    
    def start(self):
        '''
        Just set the callback function
        '''
        self.client.set_callback(
            lambda rb, s, m, t: self._update(rb, s, m, t))

    def stop(self):
        pass
    
    def get(self):
        '''
        Main logic -- parse the motion tracking data into a defined datatype
        '''

        # Run the client to collect a frame of data
        self.client.run_once(timeout=1)
        
        # Extract coordinates from feature
        coords = np.empty((self.n_features, 3))
        coords[:] = np.nan
        if self.feature == "rigid body":
            for i in range(np.min((self.n_features, len(self.rigid_bodies)))):
                if self.rigid_bodies[i].tracking_valid:
                    coords[i] = self.rigid_bodies[i].position
        elif self.feature == "marker":
            for i in range(np.min((self.n_features, len(self.markers)))):
                coords[i] = self.markers[i].position
        elif self.feature == "skeleton":
            raise NotImplementedError()
        else:
            raise AttributeError("Feature type unknown!")

        # For HDFWriter we need a dim 0
        coords = np.expand_dims(coords, axis=0)
        return coords
    
    def _update(self, rigid_bodies, skeletons, markers, timing):
        '''
        Callback for natnet client
        '''
        self.rigid_bodies = rigid_bodies
        self.skeletons = skeletons
        self.markers = markers
        self.timing = timing


#################
# Simulated data
#################
class RigidBody():

    position = None
    tracking_valid = True
    def __init__(self, position):
        self.position = position

class SimulatedClient():

    def __init__(self, n=1, radius=(0.2,0.04,0.1), speed=(0.5,1,2)):
        self.stime = time.time()
        self.n = n
        self.radius = radius
        self.speed = speed

    def set_callback(self, callback):
        self.callback = callback

    def run_once(self, timeout=None):
        '''
        Fake some motion data
        '''
        time.sleep(1./240)
        ts = (time.time() - self.stime)
        coords = np.multiply(self.radius, np.cos(np.divide(ts, self.speed) * 2 * np.pi))
        data = [RigidBody(coords)]
        self.callback(data, [], [], [])

    def start_recording(self):
        print("Start recording")

    def stop_recording(self):
        print("Stop recording")

    def set_take(self, take_name):
        print("Setting take_name: " + take_name)

    def set_session(self, session_name):
        print("Setting session_name: " + session_name)

########################
# Playback from csv file
########################

class PlaybackClient(SimulatedClient):

    def __init__(self, filename):
        import pandas as pd
        self.stime = time.time()
        csv = pd.read_csv(filename, header=[1,4,5])
        self.motiondata = csv['Rigid Body']['Position']
        self.time = csv['Type'].iloc[:,0]

    def run_once(self, timeout=None):
        '''
        Read one line of motion data from the csv file
        '''
        read_freq = 240 # doesn't really matter if we read too fast... 
        time.sleep(1./read_freq)
        ts = (time.time() - self.stime)
        coords = np.empty((3,))
        coords[:] = np.nan
        now = (i for i,t in enumerate(self.time) if t > ts) # ...because we check the timestamps here
        try:
            row = next(now)
            coords[0] = self.motiondata.iloc[row].X
            coords[1] = self.motiondata.iloc[row].Y
            coords[2] = self.motiondata.iloc[row].Z
        except:
            pass
        data = [RigidBody(coords)]
        self.callback(data, [], [], [])

# System definition function
def make(cls, client, feature, num_features=1, **kwargs):
    """
    This ridiculous function dynamically creates a class with a new init function
    """
    def init(self):
        super(self.__class__, self).__init__(client, feature, num_features, **kwargs)
    
    dtype = np.dtype((float, (num_features, 3)))
    return type(cls.__name__, (cls,), dict(dtype=dtype, __init__=init))


'''
Virtual target capture tasks where cursors are controlled by physical
motion interfaces such as joysticks
'''
import numpy as np
from scipy.spatial.transform import Rotation as R

from riglib.experiment import traits

from .target_graphics import *
from .target_capture_task import ScreenTargetCapture, ScreenReachAngle, SequenceCapture, HandConstrainedEyeCapture, ScreenTargetCapture_Saccade
from .ready_set_go_centerout import ScreenTargetCapture_ReadySet
from .target_tracking_task import ScreenTargetTracking


rotations = dict(
    yzx = np.array(    # names come from rows (optitrack), but screen coords come from columns:
        [[0, 1, 0, 0], # x goes into second column (y-coordinate, coming out of screen)
        [0, 0, 1, 0],  # y goes into third column (z-coordinate, up)
        [1, 0, 0, 0],  # z goes into first column (x-coordinate, right)
        [0, 0, 0, 1]]
    ),
    zyx = np.array(
        [[0, 0, 1, 0], 
        [0, 1, 0, 0], 
        [1, 0, 0, 0], 
        [0, 0, 0, 1]]
    ),
    xzy = np.array(
        [[1, 0, 0, 0],
        [0, 0, 1, 0], 
        [0, 1, 0, 0], 
        [0, 0, 0, 1]]
    ),
    xyz = np.identity(4),
)

baseline_rotations = dict(
    none = np.identity(4), # vertical workspace
    horizontal_workspace = np.array(
        [[1, 0, 0, 0], 
        [0, 0, 1, 0], 
        [0, 1, 0, 0], 
        [0, 0, 0, 1]]
    ),    
)

exp_rotations = dict(
    none = np.identity(4),
    about_x_90 = np.array(
        [[1, 0, 0, 0], 
        [0, 0, 1, 0], 
        [0, 1, 0, 0], 
        [0, 0, 0, 1]]
    ),
    about_x_minus_90 = np.array(
        [[1, 0, 0, 0], 
        [0, 0, -1, 0], 
        [0, 1, 0, 0], 
        [0, 0, 0, 1]]
    ),
    oop_xy_45 = np.array(
        [[ 0.707,  0.5  ,  0.5  , 0.],
         [ 0.   ,  0.707, -0.707, 0.],
         [-0.707,  0.5  ,  0.5  , 0.],
         [ 0.,     0.   ,  0.,    1.]]
    ),
    oop_xy_minus_45 = np.array(
        [[ 0.707,  0.5  , -0.5  , 0.],
         [ 0.   ,  0.707,  0.707, 0.],
         [ 0.707, -0.5  ,  0.5  , 0.],
         [ 0.,     0.   ,  0.,    1.]]
    ),
    oop_xy_20 = np.array(
        [[ 0.94 ,  0.117,  0.321, 0.],
         [-0.   ,  0.94 , -0.342, 0.],
         [-0.342,  0.321,  0.883, 0.],
         [ 0.,     0.   ,  0.,    1.]]
    ),
    oop_xy_minus_20 = np.array(
        [[ 0.94 ,  0.117, -0.321, 0.],
         [-0.   ,  0.94 ,  0.342, 0.],
         [ 0.342, -0.321,  0.883, 0.],
         [ 0.,     0.   ,  0.,    1.]]
    ),
    upright_xz_45 = np.array(
        [[ 0.707, 0.5,   0.5,   0 ],
        [-0.707, 0.5,   0.5 ,   0 ],
        [ 0.,   -0.707, 0.707 , 0 ],
        [ 0.,    0.,    0.,     1 ]]
    ),
    flat_xz_45 = np.array(
        [[ 0.707, 0.5 ,  0.5 , 0. ],
        [-0.707 ,0.5  , 0.5  , 0. ],
        [ 0.   ,-0.707, 0.707, 0. ],
        [ 0.   , 0.   , 0.   , 1. ]]
    ),
)

class ManualControlMixin(traits.HasTraits):
    '''Target capture task where the subject operates a joystick
    to control a cursor. Targets are captured by having the cursor
    dwell in the screen target for the allotted time'''

    # Settable Traits
    wait_time = traits.Float(2., desc="Time between successful trials")
    velocity_control = traits.Bool(False, desc="Position or velocity control")
    random_rewards = traits.Bool(False, desc="Add randomness to reward")
    rotation = traits.OptionsList(*rotations, desc="Rotation to transform raw to bmi3d coordinates", bmi3d_input_options=list(rotations.keys()))
    scale = traits.Float(1.0, desc="Scale factor to transform raw to bmi3d coordinates")
    baseline_rotation = traits.OptionsList(*baseline_rotations, desc="Rotation to define the mapping between bmi3d coordinates and baseline workspace", bmi3d_input_options=list(baseline_rotations.keys()))
    exp_rotation = traits.OptionsList(*exp_rotations, desc="Experimental rotation to manipulate the mapping between baseline workspace and screen coordinates", bmi3d_input_options=list(exp_rotations.keys()))
    pertubation_rotation = traits.Float(0.0, desc="Experimental rotation about bmi3d y-axis in degrees") # this is perturbation_rotation_y
    perturbation_rotation_z = traits.Float(0.0, desc="Experimental rotation about bmi3d z-axis in degrees")
    perturbation_rotation_x = traits.Float(0.0, desc="Experimental rotation about bmi3d x-axis in degrees")
    offset = traits.Array(value=[0,0,0], desc="Offset to transform raw to bmi3d coordinates")
    exp_gain = traits.Float(1.0, desc="Experimental gain scale factor to manipulate the mapping to screen coordinates")
    is_bmi_seed = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_pt=np.zeros([3]) #keep track of current pt
        self.last_pt = np.array(self.starting_pos) #keep track of last pt to calc. velocity
        self._quality_window_size = 500 # how many cycles to accumulate quality statistics
        self.reportstats['Input quality'] = "100 %"
        if self.random_rewards:
            self.reward_time_base = self.reward_time

    def init(self):
        self.add_dtype('manual_input', 'f8', (3,))
        super().init()
        self.no_data_counter = np.zeros((self._quality_window_size,), dtype='?')

    def _test_start_trial(self, ts):
        return ts > self.wait_time and not self.pause

    # def _test_trial_complete(self, ts):
    #     if self.target_index==self.chain_length-1 :
    #         if self.random_rewards:
    #             if not self.rand_reward_set_flag: #reward time has not been set for this iteration
    #                 self.reward_time = np.max([2*(np.random.rand()-0.5) + self.reward_time_base, self.reward_time_base/2]) #set randomly with min of base / 2
    #                 self.rand_reward_set_flag =1
    #                 #print self.reward_time, self.rand_reward_set_flag
    #         return self.target_index==self.chain_length-1

    def _test_reward_end(self, ts):
        #When finished reward, reset flag.
        if self.random_rewards:
            if ts > self.reward_time:
                self.rand_reward_set_flag = 0
                #print self.reward_time, self.rand_reward_set_flag, ts
        return ts > self.reward_time

    def _transform_coords(self, coords):
        ''' 
        Returns transformed coordinates based on rotation, offset, and scale traits
        '''
        offset = np.array(
            [[1, 0, 0, 0], 
            [0, 1, 0, 0], 
            [0, 0, 1, 0], 
            [self.offset[0], self.offset[1], self.offset[2], 1]]
        )
        scale = np.array(
            [[self.scale, 0, 0, 0], 
            [0, self.scale, 0, 0], 
            [0, 0, self.scale, 0], 
            [0, 0, 0, 1]]
        )
        old = np.concatenate((np.reshape(coords, -1), [1])) # manual input (3,) plus offset term
        new = np.linalg.multi_dot((old, offset, scale, rotations[self.rotation], baseline_rotations[self.baseline_rotation], exp_rotations[self.exp_rotation])) # screen coords (3,) plus offset term
        pertubation_rot = R.from_euler('y', self.pertubation_rotation, degrees=True) # this is perturb_rot_y
        perturb_rot_z = R.from_euler('z', self.perturbation_rotation_z, degrees=True)
        perturb_rot_x = R.from_euler('x', self.perturbation_rotation_x, degrees=True)
        return np.linalg.multi_dot((new[0:3] * self.exp_gain, pertubation_rot.as_matrix(), perturb_rot_z.as_matrix(), perturb_rot_x.as_matrix()))

    def _get_manual_position(self):
        '''
        Fetches joystick position
        '''
        if not hasattr(self, 'joystick'):
            return
        pt = self.joystick.get()
        if len(pt) == 0:
            return

        pt = pt[-1] # Use only the latest coordinate

        if len(pt) == 2:
            pt = np.concatenate((np.reshape(pt, -1), [0]))

        return [pt]

    def move_effector(self, pos_offset=[0,0,0], vel_offset=[0,0,0]):
        ''' 
        Sets the 3D coordinates of the cursor. For manual control, uses
        motiontracker / joystick / mouse data. If no data available, returns None
        '''

        # Get raw input and save it as task data
        raw_coords = self._get_manual_position() # array of [3x1] arrays

        if raw_coords is None or len(raw_coords) < 1:
            self.no_data_counter[self.cycle_count % self._quality_window_size] = 1
            self.update_report_stats()
            self.task_data['manual_input'] = np.ones((3,))*np.nan
            return

        self.task_data['manual_input'] = raw_coords.copy()
        self.no_data_counter[self.cycle_count % self._quality_window_size] = 0

        # Transform coordinates
        coords = self._transform_coords(raw_coords)
        
        try:
            if self.limit2d:
                coords[1] = 0
            if self.limit1d:
                coords[1] = 0
                coords[0] = 0
        except:
            if self.limit2d:
                coords[1] = 0

        # Add cursor disturbance
        coords += pos_offset
        coords += vel_offset
   
        # Set cursor position
        if not self.velocity_control:
            self.current_pt = coords
        else: # TODO how to implement velocity control?
            epsilon = 2*(10**-2) # Define epsilon to stabilize cursor movement
            if sum((coords)**2) > epsilon:

                # Add the velocity (units/s) to the position (units)
                self.current_pt = coords / self.fps + self.last_pt
            else:
                self.current_pt = self.last_pt

        self.plant.set_endpoint_pos(self.current_pt)
        self.last_pt = self.plant.get_endpoint_pos()

    def update_report_stats(self):
        super().update_report_stats()
        window_size = min(max(1, self.cycle_count), self._quality_window_size)
        num_missing = np.sum(self.no_data_counter[:window_size])
        quality = 1 - num_missing / window_size
        self.reportstats['Input quality'] = "{} %".format(int(100*quality))

    @classmethod
    def get_desc(cls, params, log_summary):
        duration = round(log_summary['runtime'] / 60, 1)
        return "{}/{} succesful trials in {} min".format(
            log_summary['n_success_trials'], log_summary['n_trials'], duration)


class ManualControl(ManualControlMixin, ScreenTargetCapture):
    '''
    Slightly refactored original manual control task
    '''
    pass

class ManualControlDirectionConstraint(ManualControlMixin, ScreenReachAngle):
    '''
    Adds an additional constraint that the direction of travel must be within a certain angle
    '''
    pass

class TrackingTask(ManualControlMixin, ScreenTargetTracking):
    '''
    Track moving target task
    '''
    pass

class SequenceTask(ManualControlMixin, SequenceCapture):
    '''
    Sequence task. Subjects need to capture 3 targets. The last target appears while reaching to the 2st target.
    '''
    pass

class HandConstrainedSaccadeTask(ManualControlMixin, HandConstrainedEyeCapture):
    '''
    Saccade task while holding different targets by hand
    '''
    pass

class SaccadeTask(ManualControlMixin, ScreenTargetCapture_Saccade):
    '''
    Center out saccade task. The controller for the cursor is eye positions. The target color changes when subjects fixate the target.
    '''
    pass

class ReadySetGoTask(ManualControlMixin, ScreenTargetCapture_ReadySet):
    '''
    Center out task with ready set go auditory cues. Cues separated by 500 ms and participant is expected to move on final go cue. Additionally, participant must move out
    of center circle (mustmv_time) parameter or there will be an error. 
    '''
    pass


'''
Handlers for AJAX (Javascript) functions used in the web interface to start
experiments and train BMI decoders
'''
import json, datetime
import logging
import io, traceback
import numpy as np
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.db.models import ProtectedError

from riglib import experiment
from .json_param import Parameters
from .models import TaskEntry, Feature, Sequence, Task, Generator, Subject, Experimenter, DataFile, System, Decoder, KeyValueStore, import_by_path
from .tasktrack import Track
from config.rig_defaults import rig_settings
import logging
import io, traceback

from . import exp_tracker # Wrapper for tasktrack.Track
from . import trainbmi

http_request_queue = []

@csrf_exempt
def train_decoder_ajax_handler(request, idx):
    '''
    AJAX handler for creating a new decoder.

    Parameters
    ----------
    request : Django HttpRequest
        POST data containing details for how to train the decoder (type, units, update rate, etc.)
    idx : int
        ID number of the TaskEntry record with the data used to train the Decoder.

    Returns
    -------
    Django HttpResponse
        Indicates 'success' if all commands initiated without error.
    '''
    ## Check if the name of the decoder is already taken
    collide = Decoder.objects.filter(entry=idx, name=request.POST['bminame'])
    if len(collide) > 0:
        return _respond(dict(status='error', msg='Name collision -- please choose a different name'))
    update_rate = float(request.POST['bmiupdaterate'])

    kwargs = dict(
        entry=idx,
        name=request.POST['bminame'],
        clsname=request.POST['bmiclass'],
        extractorname=request.POST['bmiextractor'],
        cells=request.POST['cells'],
        channels=request.POST['channels'],
        binlen=1./update_rate,
        tslice=list(map(float, request.POST.getlist('tslice[]'))),
        ssm=request.POST['ssm'],
        pos_key=request.POST['pos_key'],
        kin_extractor=request.POST['kin_extractor'],
        zscore=request.POST['zscore'],
    )
    trainbmi.cache_and_train(**kwargs)
    return _respond(dict(status="success"))


class encoder(json.JSONEncoder):
    '''
    Encoder for JSON data that defines how the data should be returned.
    '''
    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        elif isinstance(o, Parameters):
            return o.params
        else:
            return super(encoder, self).default(o)

def _respond(data):
    '''
    Generic HTTPResponse to return JSON-formatted dictionary values

    Parameters
    ----------
    data : dict
        Keys and values can be just about anything

    Returns
    -------
    HttpResponse
        JSON-encoded version of the input dictionary
    '''
    return HttpResponse(json.dumps(data, cls=encoder), content_type="application/json")

def task_info(request, idx, dbname='default'):
    '''
    Get information about the task

    Parameters
    ----------
    request : Django HttpRequest

    idx : int
        Primary key used to look up the task from the database

    Returns
    -------
    JSON-encoded dictionary
    '''
    task = Task.objects.using(dbname).get(pk=idx)
    feats = []
    for name, isset in list(request.GET.items()):
        if isset == "true": # box for the feature checked
            feat = Feature.objects.using(dbname).get(name=name)
            feats.append(feat)

    filter_kwargs = {'template': True, 'task__id': idx}
    templates = TaskEntry.objects.using(dbname).filter(**filter_kwargs).order_by("-date")
    template_info = [{'id': t.id, 'name': t.entry_name} for t in templates]

    metadata = TaskEntry.get_default_metadata()

    task_info = dict(params=task.params(feats=feats), generators=task.get_generators(), \
        templates=template_info, metadata=metadata)

    task_cls = task.get(feats=feats)
    if issubclass(task_cls, experiment.Sequence):
        task_info['sequence'] = task.sequences()
    task_info['controls'] = task.controls(feats=feats)

    return _respond(task_info)

def exp_info(request, idx, dbname='default'):
    '''
    Get information about the tasks that have already run

    Parameters
    ----------
    request : Django HttpRequest
        POST request triggered by clicking on a task entry from the left side pane
    idx : int
        Primary key used to look up the TaskEntry from the database

    Returns
    -------
    JSON-encoded dictionary
        Data containing features, parameters, and any report data from the TaskEntry
    '''
    entry = TaskEntry.objects.using(dbname).get(pk=idx)
    try:
        entry_data = entry.to_json()
    except Exception as e:
        print("##### Error trying to access task entry data: id=%s, dbname=%s" % (idx, dbname))
        import traceback
        exception = traceback.format_exc()
        exception.replace('\n', '\n    ')
        print(exception.rstrip())
        print("#####")
        return _respond_err(exception)
    else:
        return _respond(entry_data)

@csrf_exempt
def add_sequence(request):
    
    print(request.POST)
    sequence = json.loads(request.POST['sequence'])
    task_id = json.loads(request.POST.get('task'))
    seq = Sequence.from_json(sequence)
    task =  Task.objects.get(pk=task_id)
    seq.task = task
    seq.save()
    
    return _respond(dict(id=seq.id, name=seq.name))   

def hide_entry(request, idx):
    '''
    See documentation for exp_info
    '''
    print("hide_entry")
    entry = TaskEntry.objects.get(pk=idx)
    entry.visible = False
    entry.save()
    return _respond(dict())

def show_entry(request, idx):
    '''
    See documentation for exp_info
    '''
    print("hide_entry")
    entry = TaskEntry.objects.get(pk=idx)
    entry.visible = True
    entry.save()
    return _respond(dict())

def remove_entry(request, idx):
    print("Remove entry %d" % idx)
    entry = TaskEntry.objects.get(pk=idx)
    try:
        DataFile.objects.filter(entry=entry.id).delete()
    except DataFile.DoesNotExist:
        pass
    try:
        Decoder.objects.filter(entry=entry.id).delete()
    except Decoder.DoesNotExist:
        pass
    entry.delete()
    return _respond(dict())

def template_entry(request, idx):
    '''
    See documentation for exp_info
    '''
    entry = TaskEntry.objects.get(pk=idx)
    entry.template = True
    entry.save()
    return _respond(dict())

def untemplate_entry(request, idx):
    '''
    See documentation for exp_info
    '''
    entry = TaskEntry.objects.get(pk=idx)
    entry.template = False
    entry.save()
    return _respond(dict())
    
def backup_entry(request, idx):
    '''
    See documentation for exp_info
    '''
    entry = TaskEntry.objects.get(pk=idx)
    entry.backup = True
    entry.save()
    return _respond(dict())

def unbackup_entry(request, idx):
    '''
    See documentation for exp_info
    '''
    entry = TaskEntry.objects.get(pk=idx)
    entry.backup = False
    entry.save()
    return _respond(dict())

def gen_info(request, idx):
    try:
        gen = Generator.objects.get(pk=idx)
        return _respond(gen.to_json())
    except:
        traceback.print_exc()

def start_next_exp(request):
    try:
        req, save = http_request_queue.pop(0)
        return start_experiment(req, save=save)
    except IndexError:
        return _respond(dict(status="error", msg="No experiments in queue!"))

@csrf_exempt
def start_experiment(request, save=True, execute=True):
    '''
    Handles presses of the 'Start Experiment' and 'Test' buttons in the browser
    interface
    '''
    #make sure we don't have an already-running experiment
    tracker = Track.get_instance()
    if len(tracker.status.value) != 0:
        print("Task is running, exp_tracker.status.value:", tracker.status.value)
        return _respond(dict(status="running", msg="Already running task!"))

    # Try to start the task, and if there are any errors, send them to the browser interface
    try:
        data = json.loads(request.POST['data'])

        task =  Task.objects.get(pk=data['task'])
        feature_names = list(data['feats'].keys())
        subject_name = data['metadata'].pop('subject')
        subject = Subject.objects.get(name=subject_name)
        experimenter_name = data['metadata'].pop('experimenter')
        experimenter = Experimenter.objects.get(name=experimenter_name)
        project = data['metadata'].pop('project')
        session = data['metadata'].pop('session')

        entry = TaskEntry.objects.create(rig_name=rig_settings['name'], subject_id=subject.id, task_id=task.id, experimenter_id=experimenter.id,
            project=project, session=session)
        if 'entry_name' in data:
            entry.entry_name = data['entry_name']
        if 'date' in data and data['date'] != "Today" and len(data['date'].split("-")) == 3:
            datestr = data['date'].split("-")
            print("Got custom date: ", datestr)
            entry.date = datetime.datetime(int(datestr[0]), int(datestr[1]), int(datestr[2])) # this does not work: datetime.datetime.strptime("%Y-%m-%d", datetime.datetime.now().strftime("%Y-%m-%d"))

        params = Parameters.from_html(data['params'])
        entry.params = params.to_json()
        feats = Feature.getall(feature_names)
        kwargs = dict(subj=entry.subject.id, subject_name=subject_name, base_class=task.get(),
            feats=feats, params=params)
        metadata = Parameters.from_html(data['metadata'])
        entry.metadata = metadata.to_json()

        # Save the target sequence to the database and link to the task entry, if the task type uses target sequences
        if issubclass(task.get(feats=feature_names), experiment.Sequence):
            seq = Sequence.from_json(data['sequence'])
            seq.task = task
            if save:
                seq.save()
            entry.sequence = seq
            kwargs['seq'] = seq

        response = dict(status="testing", rig_name=rig_settings['name'], subj=entry.subject.name,
                        task=entry.task.name)

        if save:
            # tag software version using the git hash
            import git
            repo = git.repo.Repo(__file__, search_parent_directories=True)
            sw_version = repo.commit().hexsha[:8]
            repo_dirty = repo.is_dirty(index=True, working_tree=True, untracked_files=False)
            if repo_dirty:
                sw_version += '.dirty'
            entry.sw_version = sw_version

            # Save the task entry to database
            entry.save()

            # Link the features used to the task entry
            for feat_name in feature_names:
                f = Feature.objects.get(name=feat_name)
                entry.feats.add(f.pk)

            response['date'] = entry.date.strftime("%h %d, %Y %I:%M %p")
            response['status'] = "running"
            response['idx'] = entry.ui_id

            # Give the entry ID to the runtask as a kwarg so that files can be linked after the task is done
            kwargs['saveid'] = entry.id
        else:
            entry.delete()

        # Start the task FSM and tracker
        if execute:
            tracker.runtask(**kwargs)
        else:
            response["status"] = "completed"

        # Return the JSON response
        return _respond(response)

    except Exception as e:
        # Generate an HTML response with the traceback of any exceptions thrown
        import io
        import traceback
        from .tasktrack import log_str
        err = io.StringIO()
        traceback.print_exc(None, err)
        traceback.print_exc() # print again to console
        err.seek(0)
        log_str(err.read()) # log to tasktracker
        err.seek(0)
        tracker.reset() # make sure task is stopped
        return _respond(dict(status="error", msg=err.read()))

def rpc(fn):
    '''
    Generic remote procedure call function

    Parameters
    ----------
    fn : callable
        Function which takes a single argument, the tracker object.
        Return values from this function are ignored.

    Returns
    -------
    JSON-encoded dictionary
    '''
    tracker = Track.get_instance()

    # make sure that there exists an experiment to interact with
    if tracker.status.value not in [b"running", b"testing"]:
        print("Task not running!", str(tracker.status.value))
        return _respond(dict(status="error", msg="No task running, so cannot run command!"))

    try:
        status = tracker.status.value.decode("utf-8")
        fn_response = fn(tracker)
        response_data = dict(status="pending", msg=status)
        if not fn_response is None:
            response_data['data'] = fn_response
        return _respond(response_data)
    except Exception as e:
        import traceback
        traceback.print_exc()
        return _respond_err(e)

def _respond_err(e):
    '''
    Default error response from server to webclient

    Parameters
    ----------
    e : Exception
        Error & traceback to convert to string format.

    Returns
    -------
    JSON-encoded dictionary
        Sets status to "error" and provides the specific error message
    '''
    err = io.StringIO()
    traceback.print_exc(None, err)
    err.seek(0)
    return _respond(dict(status="error", msg=err.read()))

@csrf_exempt
def stop_experiment(request):
    return rpc(lambda tracker: tracker.stoptask())

def enable_clda(request):
    return rpc(lambda tracker: tracker.task_proxy.enable_clda())

def disable_clda(request):
    return rpc(lambda tracker: tracker.task_proxy.disable_clda())

def set_task_attr(request, attr, value):
    '''
    Generic function to change a task attribute while the task is running.
    '''
    return rpc(lambda tracker: tracker.task_proxy.remote_set_attr(attr, value))

@csrf_exempt
def save_notes(request, idx):
    te = TaskEntry.objects.get(pk=idx)
    te.notes = request.POST['notes']
    te.save()
    return _respond(dict(status="success"))

def reward_drain(request, onoff):
    '''
    Start/stop the "drain" of a solenoid reward remotely
    This function is modified to use the reward system in Orsborn lab - check reward.py for functions
    '''
    from riglib import reward
    r = reward.open()

    if onoff == 'on':
        r.drain(600)
        print('drain on')
    else:
        print('drain off')
        r.drain_off()
    return HttpResponse('Turning reward %s' % onoff)

def populate_models(request):
    """ Database initialization code. When 'db.tracker' is imported, it goes through the database and ensures that
    1) at least one subject is present
    2) all the tasks from 'tasklist' appear in the db
    3) all the features from 'featurelist' appear in the db
    4) all the generators from all the tasks appear in the db
    """
    subjects = Subject.objects.all()
    if len(subjects) == 0:
        subj = Subject(name='testing')
        subj.save()

    for m in [Generator, System]:
        m.populate()

    return HttpResponse("Updated Tasks, features generators, and systems")

@csrf_exempt
def add_new_task(request):
    from . import models
    name, import_path = request.POST['name'], request.POST['import_path']

    #  verify import path
    if import_path == '':
        import_path = "riglib.experiment.Experiment"

    try:
        import_by_path(import_path)
    except:
        import traceback
        traceback.print_exc()
        return _respond(dict(msg="import path invalid!", status="error"))

    task = Task(name=name, import_path=import_path)
    task.save()

    # add any new generators for the task
    Generator.remove_unused()       
    Generator.populate()

    task_data = dict(id=task.id, name=task.name, import_path=task.import_path)
    return _respond(dict(msg="Added new task: %s" % task.name, status="success", data=task_data))

@csrf_exempt
def remove_task(request):
    id = request.POST.get('id')
    task = Task.objects.filter(id=id)
    try:
        entry = TaskEntry.objects.filter(task=id).values_list('id', flat=True)
    except TaskEntry.DoesNotExist:
        entry = None
    if entry is None or len(entry) == 0:
        try:
            Sequence.objects.filter(task=id).delete()
        except Sequence.DoesNotExist:
            pass
        task.delete()
        return _respond(dict(msg="Removed task", status="success"))
    else:
        return _respond(dict(msg="Couldn't remove task, experiments {0} use it.".format(list(entry)), status="error"))

@csrf_exempt
def add_new_subject(request):
    subject_name = request.POST['subject_name']
    subj = Subject(name=subject_name)
    subj.save()

    return _respond(dict(msg="Added new subject: %s" % subj.name, status="success", data=dict(id=subj.id, name=subj.name)))

@csrf_exempt
def remove_subject(request):
    id = request.POST.get('id')
    try:
        Subject.objects.filter(id=id).delete()
        return _respond(dict(msg="Removed subject", status="success"))
    except ProtectedError:
        return _respond(dict(msg="Couldn't remove subject, there must be valid experiments that use it", status="error"))

@csrf_exempt
def add_new_experimenter(request):
    exp_name = request.POST['experimenter_name']
    exp = Experimenter(name=exp_name)
    exp.save()

    return _respond(dict(msg="Added new experimenter: %s" % exp.name, status="success", data=dict(id=exp.id, name=exp.name)))

@csrf_exempt
def remove_experimenter(request):
    id = request.POST.get('id')
    try:
        Experimenter.objects.filter(id=id).delete()
        return _respond(dict(msg="Removed experimenter", status="success"))
    except ProtectedError:
        return 

@csrf_exempt
def add_new_system(request):
    sys = System(name=request.POST['name'], path=request.POST['path'],
        processor_path=request.POST['processor_path'])
    sys.save()

    system_data = dict(id=sys.id, name=sys.name)
    return _respond(dict(msg="Added new system: %s" % sys.name, status="success", data=system_data))

@csrf_exempt
def remove_system(request):
    from . import models
    id = request.POST.get('id')
    try:
        System.objects.filter(id=id).delete()
        return _respond(dict(msg="Removed system", status="success"))
    except ProtectedError:
        return _respond(dict(msg="Couldn't remove system, there must be valid experiments that use it", status="error"))

@csrf_exempt
def toggle_features(request):
    from features import built_in_features
    from . import models

    name = request.POST.get('name')
    
    # check if the feature is already installed
    existing_features = Feature.objects.filter(name=name)

    if len(existing_features) > 0:
        # disable the feature
        Feature.objects.filter(name=name).delete()
        msg = "Disabled feature: %s" % str(name)
        return _respond(dict(msg=msg, status="success"))
    elif name in built_in_features:
        import_path = built_in_features[name].__module__ + '.' + built_in_features[name].__qualname__
        feat = Feature(name=name, import_path=import_path)
        feat.save()
        msg = "Enabled built-in feature: %s" % str(feat.name)
        return _respond(dict(msg=msg, status="success", id=feat.id))
    else:
        # something is wrong
        return _respond(dict(msg="feature not valid!", status="error"))   

@csrf_exempt
def add_new_feature(request):
    from . import models
    name, import_path = request.POST['name'], request.POST['import_path']

    #  verify import path
    try:
        import_by_path(import_path)
    except:
        import traceback
        traceback.print_exc()
        return _respond(dict(msg="import path invalid!", status="error"))

    feat = Feature(name=name, import_path=import_path)
    feat.save()
    
    feature_data = dict(id=feat.id, name=feat.name, import_path=feat.import_path)
    return _respond(dict(msg="Added new feature: %s" % feat.name, status="success", data=feature_data))

@csrf_exempt
def setup_run_upkeep(request):
    # Update the list of generators
    from . import models
    Generator.populate()
    return HttpResponse("Updated generators!")

@csrf_exempt
def get_report(request):
    '''
    Get data for the report field in the frontend
    '''
    def report_fn(tracker):
        tracker.task_proxy.update_report_stats()
        reportstats = tracker.task_proxy.reportstats
        return reportstats

    return rpc(report_fn)

@csrf_exempt
def trigger_control(request):
    '''
    Trigger an action via controls on the web interface
    '''
    def control_fn(tracker):
        try:
            method = getattr(tracker.task_proxy, request.POST["control"])
            if "params" in request.POST:
                params = json.loads(request.POST.get("params"))
                print(method)
                return method(**params)
            else:
                return method()
        except Exception as e:
            traceback.print_exc()

    if "base_class" in request.POST:

        # If this is a static method, it will have a base_class
        task_id = request.POST["base_class"]
        feature_names = json.loads((request.POST['feats'])).keys()
        task = Task.objects.get(pk=task_id).get(feats=feature_names)
        try:
            fn = getattr(task, request.POST["control"])
            print(fn)
            if "params" in request.POST:
                params = json.loads(request.POST.get("params"))
                result = fn(**params)
            else:
                result = fn()
            return _respond(dict(status="success", value=result))
        except Exception as e:
            traceback.print_exc()
            return _respond_err(e)
        
    else:

        # Otherwise it is a method belonging to the active task
        return rpc(control_fn)

@csrf_exempt
def get_status(request):
    """ Send the task tracker's status back to the frontend """
    tracker = Track.get_instance()
    if tracker.task_kwargs is None:
        saveid = None
    else:
        saveid = tracker.task_kwargs["saveid"]
    print("saveid", saveid)
    return _respond(dict(status=tracker.get_status(), saveid=saveid))

@csrf_exempt
def save_entry_name(request):
    from . import models
    te_rec = TaskEntry.objects.get(id=request.POST["id"])
    te_rec.entry_name = request.POST["entry_name"]
    te_rec.save()
    return _respond(dict(status="success", msg="Saved entry name: %s" % te_rec.entry_name))

def update_built_in_feature_import_paths(request):
    """For built-in features, update the import path based on the features module"""
    from . import models
    for feat in Feature.objects.all():
        feat.get(update_builtin=True)
    return _respond(dict(status="success", msg="Updated built-in feature paths!"))

def update_database_storage_path(request):
    from . import models
    db_name = request.POST['db_name']
    db_storage_path = request.POST['db_storage_path']

    KeyValueStore.set("data_path", db_storage_path, dbname=db_name)
    return _respond(dict(status="success", msg="Updated storage path for %s db" % db_name))

def save_recording_sys(request):
    from . import models
    KeyValueStore.set('recording_sys', request.POST['selected_recording_sys'])
    print(KeyValueStore.get('recording_sys'))
    ret_msg = "Set recording_sys to %s" % KeyValueStore.get('recording_sys')
    return _respond(dict(status="success", msg=ret_msg))


@csrf_exempt
def setup_handler(request):
    """One-stop handler for setup functions to avoid adding a bunch of URLs"""
    action = request.POST['action']
    if action == "update_database_storage_path":
        return update_database_storage_path(request)
    elif action == "save_recording_sys":
        return save_recording_sys(request)
    elif action == "update_built_in_feature_paths":
        return update_built_in_feature_import_paths(request)
    else:
        return _respond(dict(status="error", msg="Unrecognized data type: %s" % action))




'''
FBO (Framebuffer Object) Module

An FBO is an OpenGL object that allows rendering to off-screen buffers, which can be used as textures.
This is useful for various rendering techniques like post-processing effects, shadow mapping, or
rendering to multiple targets simultaneously.

Key components:
- FBO: The main framebuffer object
- Attachments: Textures or renderbuffers that store the rendered output (color, depth, stencil)
- Render targets: The buffers that OpenGL draws into (color attachments, depth attachment, etc.)

This module provides classes to create and manage FBOs, as well as utilities for rendering to them.
'''

import numpy as np
from OpenGL.GL import *

from .render import Renderer
from ..textures import Texture

fbotypes = dict(
    depth=(GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT, GL_FLOAT, GL_DEPTH_ATTACHMENT), 
    stencil=(GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, GL_STENCIL_ATTACHMENT), 
    colors=(GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE, GL_COLOR_ATTACHMENT0)
)

class FBO(object):
    def __init__(self, attachments, size=None, ncolors=1, **kwargs):
        maxcolors = range(glGetInteger(GL_MAX_COLOR_ATTACHMENTS))
        self.names = dict(("color%d"%i, GL_COLOR_ATTACHMENT0+i) for i in maxcolors)
        self.names["depth"] = GL_DEPTH_ATTACHMENT
        self.names["stencil"] = GL_STENCIL_ATTACHMENT

        self.textures = dict()

        self.fbo = glGenFramebuffers(1)
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        
        for attach in attachments:
            if isinstance(attach, str):
                if attach.startswith("color"):
                    idx = int(attach[5:])
                    attach = "colors"
                iform, exform, dtype, attachment = fbotypes[attach]
                texture = Texture(None, size=size, iformat=iform, exformat=exform, dtype=dtype)
                texture.init()
                if attach == "colors":
                    attachment += idx
                glFramebufferTexture2D(GL_FRAMEBUFFER, attachment, GL_TEXTURE_2D, texture.tex, 0)
                self.textures[attachment] = texture
            else:
                attachment, texture = attach
                if attachment in self.names:
                    attachment = self.names[attachment]
                if texture is None and attachment == GL_DEPTH_ATTACHMENT:
                    rb = glGenRenderbuffers(1)
                    glBindRenderbuffer(GL_RENDERBUFFER, rb)
                    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, int(size[0]), int(size[1]))
                    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, rb)
                else:
                    if texture.tex is None:
                        texture.init()
                    glFramebufferTexture2D(GL_FRAMEBUFFER, attachment, GL_TEXTURE_2D, texture.tex, 0)
                    self.textures[attachment] = texture
                
        types = [t for t in list(self.textures.keys()) if 
            t !=GL_DEPTH_ATTACHMENT and 
            t != GL_STENCIL_ATTACHMENT and 
            t != GL_DEPTH_STENCIL_ATTACHMENT]
        if len(types) > 0:
            glDrawBuffers(types)
        else:
            glDrawBuffers(GL_NONE)
        
        assert glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE
        glBindFramebuffer(GL_FRAMEBUFFER, 0)
    
    def __getitem__(self, idx):
        if isinstance(idx, str):
            return self.textures[self.names[idx]]
        
        return self.textures[idx]
    
    def clear(self):
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)

class FBOrender(Renderer):

    def __init__(self, *args, **kwargs):
        super(FBOrender, self).__init__(*args, **kwargs)
        self.vao = glGenVertexArrays(1)

    def draw_fsquad(self, shader, **kwargs):
        ctx = self.programs[shader]
        glUseProgram(ctx.program)
        for name, v in list(kwargs.items()):
            if isinstance(v, Texture):
                ctx.uniforms[name] = self.get_texunit(v)
            else:
                ctx.uniforms[name] = v
        
        glBindVertexArray(self.vao)
        glEnableVertexAttribArray(ctx.attributes['position'])
        glBindBuffer(GL_ARRAY_BUFFER, self.fsquad_buf[0])
        glVertexAttribPointer(ctx.attributes['position'],
            4, GL_FLOAT, GL_FALSE, 4*4, GLvoidp(0))
        
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.fsquad_buf[1]);
        glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, GLvoidp(0))
        glDisableVertexAttribArray(ctx.attributes['position'])
        glBindVertexArray(0)

    def draw_fsquad_to_fbo(self, fbo, shader, **kwargs):
        glBindFramebuffer(GL_FRAMEBUFFER, fbo.fbo)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)
        self.draw_fsquad(shader, **kwargs)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)

    def draw_to_fbo(self, fbo, root, **kwargs):
        glBindFramebuffer(GL_FRAMEBUFFER, fbo.fbo)
        #Erase old buffer info
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)
        super(FBOrender, self).draw(root, **kwargs)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)
    